import numpy as np
from aegis_sim import constants

from aegis_sim.submodels.genetics.composite.interpreter import Interpreter
from aegis_sim import parameterization


class CompositeArchitecture:
    """

    GUI
    - when pleiotropy is not needed;
    - it is quick, easy to analyze, delivers a diversity of phenotypes
    - every trait (surv repr muta neut) can be evolvable or not
    - if not evolvable, the value is set by !!!
    - if evolvable, it can be agespecific or age-independent
    - probability of a trait at each age is determined by a BITS_PER_LOCUS adjacent bits forming a "locus" / gene
    - the method by which these loci are converted into a phenotypic value is the Interpreter type

    """

    def __init__(self, ploid, BITS_PER_LOCUS, AGE_LIMIT, THRESHOLD):
        self.ploid = ploid
        self.BITS_PER_LOCUS = BITS_PER_LOCUS
        self.n_loci = constants.TRAIT_N * AGE_LIMIT
        self.length = self.n_loci * BITS_PER_LOCUS
        self.AGE_LIMIT = AGE_LIMIT

        self.evolvable = [trait for trait in parameterization.traits.values() if trait.evolvable]

        self.interpreter = Interpreter(
            self.BITS_PER_LOCUS,
            THRESHOLD,
        )

    def get_number_of_phenotypic_values(self):
        return self.AGE_LIMIT * constants.TRAIT_N

    def get_number_of_bits(self):
        return self.ploid.y * self.n_loci * self.BITS_PER_LOCUS

    def get_shape(self):
        return (self.ploid.y, self.n_loci, self.BITS_PER_LOCUS)

    def init_genome_array(self, popsize):
        # TODO enable initgeno
        # TODO enable agespecific False
        array = np.random.random(size=(popsize, *self.get_shape()))

        for trait in parameterization.traits.values():
            array[:, :, trait.slice] = array[:, :, trait.slice] < trait.initgeno

        return array

    def init_phenotype_array(self, popsize):
        return np.zeros(shape=(popsize, self.get_number_of_phenotypic_values()))

    def compute(self, genomes):

        if genomes.shape[1] == 1:  # Do not calculate mean if genomes are haploid
            genomes = genomes[:, 0]
        else:
            genomes = self.ploid.diploid_to_haploid(genomes)

        interpretome = np.zeros(shape=(genomes.shape[0], genomes.shape[1]), dtype=np.float32)
        for trait in parameterization.traits.values():
            loci = genomes[:, trait.slice]  # fetch
            probs = self.interpreter.call(loci, trait.interpreter)  # interpret
            # self.diffuse(probs)
            interpretome[:, trait.slice] += probs  # add back

        return interpretome

    # def diffuse(self, probs):
    #     window_size = parametermanager.parameters.DIFFUSION_FACTOR * 2 + 1
    #     p = np.empty(shape=(probs.shape[0], probs.shape[1] + window_size - 1))
    #     p[:, :window_size] = np.repeat(probs[:, 0], window_size).reshape(-1, window_size)
    #     p[:, window_size - 1 :] = probs[:]
    #     diffusome = np.convolve(p[0], np.ones(window_size) / window_size, mode="valid")

    def get_map(self):
        pass



from multiprocessing import Process
import time
import pathlib
from datetime import datetime
from .recorder import Recorder
import logging


class Ticker(Recorder):
    def __init__(self, TICKER_RATE, odir: pathlib.Path):
        self.TICKER_RATE = TICKER_RATE
        self.ticker_path = odir / "ticker.txt"
        self.process = None
        self.pid = None

    def start_process(self):
        self.process = Process(target=self.tick)
        self.process.start()
        self.pid = self.process.pid
        # TODO check if this continues when simulation is interrupted or terminated otherwise

    def stop_process(self):
        self.process.terminate()
        self.process.join()

    def tick(self):
        while True:
            self.write()
            time.sleep(self.TICKER_RATE)

    def write(self):
        """
        # OUTPUT SPECIFICATION
        path: /ticker.txt
        filetype: txt
        category: log
        description: A live file useful for determining whether the simulation is still running. It gets updated every TICKER_RATE seconds; if it is not updated, the simulation is not running.
        trait granularity: N/A
        time granularity: N/A
        frequency parameter: TICKER_RATE
        structure: A txt file with datetime stamp (%Y-%m-%d %H:%M:%S) in one line)
        """
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        with open(self.ticker_path, "w") as file:
            file.write(timestamp)

    def read(self):
        if not self.ticker_path.exists():
            logging.error(f"{self.ticker_path} does not exist.")
            return
        with open(self.ticker_path, "r") as file:
            return file.read()

    def has_stopped(self):
        since_last = self.since_last()
        return since_last > self.TICKER_RATE

    def since_last(self):
        timestamp_recorded = self.read()
        if timestamp_recorded is None or timestamp_recorded == "":
            logging.info(f"timestamp_recorded is '{timestamp_recorded}'")
            return
        timestamp_now = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        dt_recorded = datetime.strptime(timestamp_recorded, "%Y-%m-%d %H:%M:%S")
        dt_now = datetime.strptime(timestamp_now, "%Y-%m-%d %H:%M:%S")
        time_difference = (dt_now - dt_recorded).total_seconds()
        return time_difference



import numpy as np


def recombination(genomes, RECOMBINATION_RATE):
    """Return recombined chromatids."""

    if RECOMBINATION_RATE == 0:
        return genomes

    # Recombine two chromatids but pass only one;
    #   thus double the number of chromatids, recobine,
    #   then return only one chromatid from each chromatid pair
    genomes = genomes[np.repeat(np.arange(len(genomes)), 2)]

    # Flatten loci and bits
    flat_genomes = genomes.reshape(len(genomes), 2, -1)

    # Get chromatids
    chromatid1 = flat_genomes[:, 0]
    chromatid2 = flat_genomes[:, 1]

    # Make choice array: when to take recombined and when to take original loci
    # -1 means synapse; +1 means clear
    rr = RECOMBINATION_RATE / 2  # / 2 because you are generating two random vectors (fwd and bkd)
    reco_fwd = (np.random.random(chromatid1.shape) < rr) * -2 + 1
    reco_bkd = (np.random.random(chromatid2.shape) < rr) * -2 + 1

    # Propagate synapse
    reco_fwd_cum = np.cumprod(reco_fwd, axis=1)
    reco_bkd_cum = np.cumprod(reco_bkd[:, ::-1], axis=1)[:, ::-1]

    # Recombine if both sites recombining
    reco_final = (reco_fwd_cum + reco_bkd_cum) == -2

    # Choose bits from first or second chromatid
    # recombined = np.empty(flat_genomes.shape, bool)
    recombined = np.empty(flat_genomes.shape, dtype=np.bool_)
    recombined[:, 0] = np.choose(reco_final, [chromatid1, chromatid2])
    recombined[:, 1] = np.choose(reco_final, [chromatid2, chromatid1])
    recombined = recombined.reshape(genomes.shape)

    return recombined[::2]  # Look at first comment in the function



import dash
from dash import callback, Output, Input, ALL, State, dcc

from aegis_gui.utilities import log
import dash_bootstrap_components as dbc


# def make_single_dropdown(dropdown_options):
#     # TODO plot initial plots
#     initial_value = dropdown_options[0] if dropdown_options else None
#     initial_value = None
#     return dcc.Dropdown(id="dropdown-single", options=dropdown_options, value=initial_value)


def make_multi_dropdown(dropdown_options, dropdown_value=[]):
    if dropdown_value == []:
        dropdown_value = [dropdown_options[0]] if dropdown_options else []
    assert isinstance(dropdown_value, list)
    return dbc.InputGroup(
        [
            # dbc.InputGroupText("Plotted simulation"),
            dcc.Dropdown(
                id="dropdown-multi",
                options=dropdown_options,
                value=dropdown_value,
                multi=True,
                placeholder="Select simulations to plot...",
                className="plot-dropdown",
                style={"width": "100%"},
                persistence=True,
                clearable=False,
            ),
        ]
    )


@callback(
    Output({"type": "loading-graph-div", "index": ALL}, "parent_style"),
    Input("figure-select", "value"),
    State({"type": "loading-graph-div", "index": ALL}, "id"),
    State({"type": "loading-graph-div", "index": ALL}, "parent_style"),
)
def change_visibility(figure_selected, graph_ids, parent_styles):

    new_parent_styles = []

    for graph_id, parent_style in zip(graph_ids, parent_styles):
        new_parent_style = parent_style.copy() if parent_style else {}
        should_be_invisible = graph_id["index"] != figure_selected

        if should_be_invisible:
            new_parent_style["display"] = "none"
        elif "display" in new_parent_style:
            del new_parent_style["display"]

        new_parent_styles.append(new_parent_style)

    return new_parent_styles



import numpy as np


class Mutator:
    def __init__(self, MUTATION_RATIO, MUTATION_METHOD, MUTATION_AGE_MULTIPLIER):
        self.MUTATION_RATIO = MUTATION_RATIO
        self.MUTATION_METHOD = MUTATION_METHOD
        self.MUTATION_AGE_MULTIPLIER = MUTATION_AGE_MULTIPLIER
        self.rate_0to1 = MUTATION_RATIO / (1 + MUTATION_RATIO)
        self.rate_1to0 = 1 / (1 + MUTATION_RATIO)
        # Set mutation method
        if self.MUTATION_METHOD == "by_index":
            self._mutate = self._mutate_by_index
        elif self.MUTATION_METHOD == "by_bit":
            self._mutate = self._mutate_by_bit
        else:
            raise ValueError("MUTATION_METHOD must be 'by_index' or 'by_bit'")

    def _mutate_by_bit(self, genomes, muta_prob, ages, random_probabilities=None):
        """Induce germline mutations."""

        if random_probabilities is None:
            random_probabilities = np.random.random(genomes.shape)

        # Broadcast to fit [individual, chromatid, locus, bit] shape
        mutation_probabilities = muta_prob[:, None, None, None]

        mutation_probabilities = self.apply_mutation_age_multiplier(
            mutation_probabilities=mutation_probabilities,
            ages=ages,
            MUTATION_AGE_MULTIPLIER=self.MUTATION_AGE_MULTIPLIER,
        )
        mutate_0to1 = (~genomes) & (
            random_probabilities < (mutation_probabilities * self.rate_0to1).astype("float32")
        )  # genome == 0 &
        mutate_1to0 = genomes & (
            random_probabilities < (mutation_probabilities * self.rate_1to0).astype("float32")
        )  # genomes == 1 &

        genomes[mutate_0to1] = 1
        genomes[mutate_1to0] = 0

        return genomes

    def _mutate_by_index(self, genomes, muta_prob, ages):
        """Alternative faster method for introducing mutations.

        Instead of generating a random probability for every bit in the array of genomes,
        generate random indices of bits that could be mutated."""

        bits_per_genome = genomes[0].size

        muta_prob = self.apply_mutation_age_multiplier(
            mutation_probabilities=muta_prob,
            ages=ages,
            MUTATION_AGE_MULTIPLIER=self.MUTATION_AGE_MULTIPLIER,
        )

        # Calculate number of bits to mutate
        n_mutations_per_individual = np.random.binomial(n=bits_per_genome, p=muta_prob, size=len(genomes))
        n_mutations_total = np.sum(n_mutations_per_individual)

        # Generate indices to mutate
        mutation_indices = (
            np.repeat(np.arange(len(genomes)), n_mutations_per_individual),
            np.random.integers(genomes.shape[1], size=n_mutations_total),
            np.random.integers(genomes.shape[2], size=n_mutations_total),
            np.random.integers(genomes.shape[3], size=n_mutations_total),
        )

        # Extract indices of 0-bits and 1-bits
        bits = genomes[mutation_indices]  # NOTE Use tuple for ndarray indexing
        bits0_indices = (~bits).nonzero()[0]
        bits1_indices = bits.nonzero()[0]

        # Take into consideration the MUTATION_RATIO
        bits0_include = np.random.random(len(bits0_indices)) < self.rate_0to1
        bits1_include = np.random.random(len(bits1_indices)) < self.rate_1to0
        bits0_indices = bits0_indices[bits0_include]
        bits1_indices = bits1_indices[bits1_include]

        # Mutate bits at mutation_indices
        mutation_indices = np.array(mutation_indices)
        genomes[tuple(mutation_indices[:, bits1_indices.T])] = False
        genomes[tuple(mutation_indices[:, bits0_indices.T])] = True

        return genomes

    @staticmethod
    def apply_mutation_age_multiplier(mutation_probabilities, ages, MUTATION_AGE_MULTIPLIER):
        """
        final = initial(1 + age * multiplier)
        """
        multipliers = ages * MUTATION_AGE_MULTIPLIER
        multipliers = multipliers.reshape(mutation_probabilities.shape)
        return mutation_probabilities * (1 + multipliers)



import dash
import dash_bootstrap_components as dbc
from aegis_gui.guisettings.GuiSettings import gui_settings


def get_offcanvas_trigger():
    return dbc.Button(
        children=[dash.html.I(className="bi bi-gear-wide"), "Preferences"],
        id="open-offcanvas-backdrop",
        n_clicks=0,
        outline=True,
        color="primary",
        style={"marginTop": "2rem"},
    )


def get_theme_switch():
    return dash.html.Div(
        [
            dbc.Label("Dark mode", html_for="color-mode-switch"),
            dbc.Switch(id="color-mode-switch", value=False, className="d-inline-block ms-1", persistence=True),
        ]
    )


def get_delete_switch():
    return dash.html.Div(
        [
            dbc.Label("One-click deletion", html_for="one-click-deletion"),
            dbc.Switch(
                id="one-click-deletion",
                value=False,
                className="d-inline-block ms-1",
                persistence=True,
            ),
        ]
    )


def get_gui_env_settings():
    return dash.html.Div(
        [
            dash.html.Hr(),
            # dash.html.P("GUI settings"),
            dash.html.P(
                f"environment: {gui_settings.ENVIRONMENT}",
                className="text-secondary",
            ),
            dash.html.P(
                f"simulation number limit: {gui_settings.SIMULATION_NUMBER_LIMIT}",
                className="text-secondary",
            ),
        ],
        style={"padding": "0 0.5rem 0.25rem 0.5rem"},
    )


def get_offcanvas():
    children = [
        get_theme_switch(),
        get_delete_switch() if gui_settings.ENVIRONMENT == "local" else None,
        get_gui_env_settings(),
    ]
    return dbc.Offcanvas(
        children=children,
        id="offcanvas-backdrop",
        scrollable=True,
        title="User preferences",
    )


@dash.callback(
    dash.Output("offcanvas-backdrop", "is_open"),
    dash.Input("open-offcanvas-backdrop", "n_clicks"),
    dash.State("offcanvas-backdrop", "is_open"),
)
def toggle_offcanvas(n1, is_open):
    if n1:
        return not is_open
    return is_open


dash.clientside_callback(
    """
    (switchOn) => {
       document.documentElement.setAttribute('data-bs-theme', switchOn ? 'dark' : 'false');  
       return window.dash_clientside.no_update
    }
    """,
    dash.Output("color-mode-switch", "id"),
    dash.Input("color-mode-switch", "value"),
)



import pathlib


class Documenter:

    here = pathlib.Path(__file__).absolute().parent

    def __init__(self):
        pass

    @staticmethod
    def read(filename):
        with open(Documenter.here / filename, "r") as file_:
            md = file_.read()
        return md



import dash
from aegis_gui.utilities import utilities, log, manipulate

import dash_bootstrap_components as dbc


def make_button(filename, disabled=False):
    return dbc.Button(
        [dash.html.I(className="bi bi-stop-circle"), "Terminate"],
        id={"type": "config-terminate-button", "index": filename},
        color="warning",
        value=filename,
        className="me-2",
        disabled=disabled,
        # TODO add disabled if there is nothing to terminate
    )


@dash.callback(
    dash.Output({"type": "config-terminate-button", "index": dash.MATCH}, "n_clicks"),
    dash.Input({"type": "config-terminate-button", "index": dash.MATCH}, "n_clicks"),
)
def button_terminate_simulation(n_clicks):
    if n_clicks is None:
        return dash.no_update

    button_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
    filename = eval(button_id)["index"]
    # TODO do not use eval
    manipulate.terminate_simulation(filename)
    return dash.no_update



"""Data recorder

Records data generated by the simulation.

When thinking about recording additional data, consider that there are three recording methods:
    I. Snapshots (record data from the population at a specific step)
    II. Flushes (collect data over time then flush)
    III. One-time records
    IV. Other: TE records
"""

import pathlib
import shutil
import logging

from aegis_sim import variables

from .terecorder import TERecorder
from .picklerecorder import PickleRecorder
from .popgenstatsrecorder import PopgenStatsRecorder
from .intervalrecorder import IntervalRecorder
from .flushrecorder import FlushRecorder
from .featherrecorder import FeatherRecorder
from .phenomaprecorder import PhenomapRecorder
from .summaryrecorder import SummaryRecorder
from .progressrecorder import ProgressRecorder
from .simpleprogressrecorder import SimpleProgressRecorder
from .popsizerecorder import PopsizeRecorder
from .ticker import Ticker
from .configrecorder import ConfigRecorder

# TODO write tests


class RecordingManager:
    """
    Container class for various recorders.
    Each recorder records a certain type of data.
    Most recorders record data as tables, except SummaryRecorder and PickleRecorder which record JSON files and pickles (a binary python format).
    Headers and indexes of all tabular files are explicitly recorded.

    -----
    GUI
    AEGIS records a lot of different data.
    In brief, AEGIS records
    genomic data (population-level allele frequencies and individual-level binary sequences) and
    phenotypic data (observed population-level phenotypes and intrinsic individual-level phenotypes),
    as well as
    derived demographic data (life, death and birth tables),
    population genetic data (e.g. effective population size, theta), and
    survival analysis data (TE / time-event tables).
    Furthermore, it records metadata (e.g. simulation log, processed configuration files) and python pickle files.

    Recorded data is distributed in multiple files.
    Almost all data are tabular, so each file is a table to which rows are appended as the simulation is running.
    The recording rates are frequencies at which rows are added; they are expressed in simulation steps.
    """

    def init(self, custom_config_path, overwrite):
        self.odir = self.make_odir(custom_config_path=custom_config_path, overwrite=overwrite)
        # TODO make subfolders

    def initialize_recorders(self, TICKER_RATE):
        self.terecorder = TERecorder(odir=self.odir)
        self.picklerecorder = PickleRecorder(odir=self.odir)
        self.popgenstatsrecorder = PopgenStatsRecorder(odir=self.odir)
        self.guirecorder = IntervalRecorder(odir=self.odir)
        self.flushrecorder = FlushRecorder(odir=self.odir)
        self.featherrecorder = FeatherRecorder(odir=self.odir)
        self.phenomaprecorder = PhenomapRecorder(odir=self.odir)
        self.summaryrecorder = SummaryRecorder(odir=self.odir)
        self.progressrecorder = ProgressRecorder(odir=self.odir)
        self.simpleprogressrecorder = SimpleProgressRecorder(odir=self.odir)
        self.ticker = Ticker(odir=self.odir, TICKER_RATE=TICKER_RATE)
        self.popsizerecorder = PopsizeRecorder(odir=self.odir)
        self.configrecorder = ConfigRecorder(odir=self.odir)

    #############
    # UTILITIES #
    #############

    @staticmethod
    def make_odir(custom_config_path, overwrite) -> pathlib.Path:
        output_path = custom_config_path.parent / custom_config_path.stem  # remove .yml
        is_occupied = output_path.exists() and output_path.is_dir()
        if is_occupied:
            if overwrite:
                shutil.rmtree(output_path)
            else:
                raise Exception(f"{output_path} already exists. To overwrite, add flag --overwrite or -o.")
        return output_path

    @staticmethod
    def make_subfolders(paths):
        # TODO relink paths that are now in classes
        for path in paths:
            path.mkdir(exist_ok=True, parents=True)

    def is_extinct(self) -> bool:
        if self.summaryrecorder.extinct:
            logging.info(f"Population went extinct (at step {variables.steps}).")
            return True
        return False



import numpy as np

from aegis_sim.utilities.funcs import skip


from .recorder import Recorder
from aegis_sim import submodels


class PopgenStatsRecorder(Recorder):
    def __init__(self, odir):
        self.odir = odir / "popgen"
        self.init_odir()

    def write(self, genomes, mutation_rates):
        """
        Record population size in popgenstats, and record popgen statistics

        # OUTPUT SPECIFICATION
        path: /popgen/simple.csv
        filetype: csv
        category: population genetics
        description: Simple population metrics including population size, effective population size, mu, segregating sites, segregating sites using a genomic sample, theta, theta_w, theta_pi, tajimas_d, theta_h, and fayandwu_h.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/allele_frequencies.csv
        filetype: csv
        category: population genetics
        description: 1-allele population-frequencies of every genomic site.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/genotype_frequencies.csv
        filetype: csv
        category: population genetics
        description: Genotype frequencies at site resolution (e.g. for a diploid genome, number of 00, 01 and 11 for each site).
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/sfs.csv
        filetype: csv
        category: population genetics
        description: A site frequency spectrum.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/mean_h_per_bit_expected.csv
        filetype: csv
        category: population genetics
        description: Heterozygosity per bit.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/mean_h_per_bit.csv
        filetype: csv
        category: population genetics
        description: Expected mean heterozygosity per bit under Hardy-Weinberg-Equilibrium.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/mean_h_per_locus.csv
        filetype: csv
        category: population genetics
        description: Mean bit heterozygosity per locus.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/reference_genome.csv
        filetype: csv
        category: population genetics
        description: Reference genome based on which allele is most common at each position.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.

        # OUTPUT SPECIFICATION
        path: /popgen/reference_genome_gsample.csv
        filetype: csv
        category: population genetics
        description: Reference genome based on which allele is most common at each position in a sample of genomes.
        trait granularity:
        time granularity:
        frequency parameter: POPGENSTATS_RATE
        structure: A float matrix.
        """
        submodels.popgenstats.record_pop_size_history(genomes.array)

        if skip("POPGENSTATS_RATE") or len(genomes) == 0:
            return

        submodels.popgenstats.calc(genomes.array, mutation_rates)

        # Record simple statistics
        array = list(submodels.popgenstats.emit_simple().values())
        if None in array:
            return

        with open(self.odir / "simple.csv", "ab") as f:
            np.savetxt(f, [array], delimiter=",", fmt="%1.3e")

        # TODO when writing some metrics (e.g. reference genome, reference genome gsample) use the appropriate dtype (bool in that case)

        # Record complex statistics
        complex_statistics = submodels.popgenstats.emit_complex()
        for key, array in complex_statistics.items():
            with open(self.odir / f"{key}.csv", "ab") as f:
                np.savetxt(f, [array], delimiter=",", fmt="%1.3e")



import numpy as np

class MatingManager:
    def __init__(self):
        pass

    def pair_up_polygamously(self, sexes):
        """
        Return indices of reproducing, sex-sorted individuals.
        Make sure no same-sex fertilization is happening.
        """

        indices_male = (sexes == 0).nonzero()[0]
        indices_female = (sexes == 1).nonzero()[0]

        # Compute number of pairs
        n_males = len(indices_male)
        n_females = len(indices_female)
        n_pairs = min(n_males, n_females)

        # Shuffle
        np.random.shuffle(indices_male)
        np.random.shuffle(indices_female)

        # Pair up
        males = indices_male[:n_pairs]
        females = indices_female[:n_pairs]

        return males, females

    def pair_up_monogamously(self, sexes):
        return



from aegis_gui.guisettings.GuiSettings import gui_settings
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS


def is_input_in_valid_range(input_, param_name: str) -> bool:
    param = DEFAULT_PARAMETERS[param_name]
    input_ = param.convert(input_)
    in_inrange = param.inrange(input_)

    if gui_settings.ENVIRONMENT == "server":
        in_serverrange = param.serverrange(input_)
    else:
        in_serverrange = True

    return in_serverrange and in_inrange



import dash
import logging

from dash import html, dcc
from aegis_gui.utilities import log, utilities
from aegis_gui.pages.tab_config import config_input, run_simulation_button, use_prerun

from aegis_gui.pages.tab_config import foldable, copy_config

from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS

from aegis_gui.docs.domains import TEXTS_DOMAIN

dash.register_page(__name__, path="/config", name="AEGIS | Launch")

# # TODO source from documentation
# TEXTS_DOMAIN = {
#     "starvation": utilities.extract_gui_from_docstring(starvation.Starvation),
#     "predation": utilities.extract_gui_from_docstring(predation.Predation),
#     "infection": utilities.extract_gui_from_docstring(infection.Infection),
#     "abiotic": utilities.extract_gui_from_docstring(abiotic.Abiotic),
#     "reproduction": utilities.extract_gui_from_docstring(Reproducer),
#     "recording": utilities.extract_gui_from_docstring(RecordingManager),
#     "genetics": """
#     Every individual carries their own genome. In AEGIS, those are bit strings (arrays of 0's and 1's), passed on from parent to offspring, and mutated in the process.
#     The submodel genetics transforms genomes into phenotypes; more specifically – into intrinsic phenotypes – biological potentials to exhibit a certain trait (e.g. probability to reproduce).
#     These potentials are either realized or not realized, depending on the environment (e.g. availability of resources), interaction with other individuals (e.g. availability of mates) and interaction with other traits (e.g. survival).

#     In AEGIS, genetics is simplified in comparison to the biological reality – it references no real genes and it simulates no molecular interactions; thus, it cannot be used to answer questions about specific genes, metabolic pathways or molecular mechanisms.
#     However, in AEGIS, in comparison to empirical datasets, genes are fully functionally characterized (in terms of their impact on the phenotype), and are to be studied as functional, heritable genetic elements – in particular, their evolutionary dynamics.

#     The configuration of genetics – the genetic architecture – is highly flexible. This includes specifying which traits are evolvable number of genetic elements (i.e. size of genome)...
#     AEGIS offers two genetic architectures – composite and modifying. They are mutually exclusive and are described in detail below...
#     """,
#     "composite genetic architecture": utilities.extract_gui_from_docstring(CompositeArchitecture),
#     "modifying genetic architecture": utilities.extract_gui_from_docstring(ModifyingArchitecture),
#     "environmental drift": utilities.extract_gui_from_docstring(Envdrift),
#     "technical": "",
#     "other": "",
# }


PREFACE = [
    html.Div(
        children=[
            # TODO change text
            html.H1("Configuration tab"),
            # html.P(
            #     [
            #         # """Customize and start simulations.
            #         # """
            #         # """
            #         # This is the configuration tab. Here you can customize parameters and start simulations.
            #         # On the left side, there are the parameter tables in which you can specify custom parameter values.
            #         # Each parameter has a valid data type and a valid range; when the input value is not valid, the simulation cannot be started.
            #         # If no input is given, the parameter will take on the default value.
            #         # Parameters are grouped into tables depending on which simulated process (submodel) they customize – these are, in brief, mortality (infection, predation, starvation, abiotic mortality), reproduction, genetics, environmental drift and recording.
            #         # Each submodel is described on the right side, next to the relevant parameter table.
            #         # To start the simulation, adjust parameter values, enter a unique ID in the input bar on top and click the adjacent button "run simulation".
            #         # """,
            #     ],
            # ),
        ]
    )
]

HEADER = html.Tr(
    [
        html.Th("PARAMETER", style={"padding-left": "1.2rem"}),
        html.Th("TYPE"),
        html.Th("RANGE", className="valid-values"),
        html.Th("VALUE"),
        # html.Th("DESCRIPTION", style={"padding-right": "1.2rem"}),
    ],
)


def layout() -> html.Div:
    # Group parameters by domain
    subsets = {domain: [] for domain in TEXTS_DOMAIN.keys()}
    for param in DEFAULT_PARAMETERS.values():
        if param.domain not in subsets:
            logging.error(f"Parameter domain {param.domain} has no documentation.")
            subsets[param.domain] = []
        subsets[param.domain].append(param)

    # Generate layout

    # tables = []
    # for domain, subset in subsets.items():
    #     tables.append(
    #         html.Div(
    #             children=[
    #                 html.P(domain, className="config-domain"),
    #                 html.Div(
    #                     children=[
    #                         html.Div(
    #                             children=[
    #                                 html.Div(
    #                                     children=[get_table(subset)],
    #                                     # style={"width": "50%"},
    #                                 ),
    #                             ],
    #                         ),
    #                         html.Div(
    #                             html.Div(children=TEXTS_DOMAIN.get(domain, ""), className="config-domain-desc"),
    #                         ),
    #                     ],
    #                     style={"display": "flex"},
    #                 ),
    #             ],
    #         )
    #     )

    foldables = foldable.get_foldable()

    # Generate layout
    return html.Div(
        id="sim-section",
        # style={"display": "none"},
        children=PREFACE
        + copy_config.make_select()
        + foldables
        + [use_prerun.make_select()]
        + [run_simulation_button.layout],
    )


# def get_row(v: Parameter) -> html.Tr:
#     if v.serverrange_info:
#         serverrange_info_message = f"Allowed parameter range for the server is {v.serverrange_info}."
#     else:
#         serverrange_info_message = ""

#     return html.Tr(
#         [
#             # PARAMETER
#             html.Td(
#                 v.get_name(),
#                 style={"padding-left": "1.2rem"},
#                 title=v.info if v.info else None,
#             ),
#             # TYPE
#             html.Td(
#                 children=html.Label(
#                     v.dtype.__name__,
#                     className=f"dtype-{v.dtype.__name__} dtype",
#                 )
#             ),
#             # RANGE
#             html.Td(children=v.drange, className="data-range"),
#             # VALUE
#             html.Td(children=config_input.get_input_element(param=v)),
#             # DESCRIPTION
#             # html.Td(
#             #     children=[
#             #         v.info + ".",
#             #         html.P(
#             #             serverrange_info_message,
#             #             className="serverrange_info_message",
#             #         ),
#             #     ],
#             #     className="td-info",
#             #     style={"padding-right": "0.8rem"},
#             # ),
#         ],
#     )


# def get_table(params_subset) -> html.Table:
#     return html.Table(
#         className="config-table",
#         children=[
#             html.Col(className="config-table-col"),
#             html.Col(className="config-table-col"),
#             html.Col(className="config-table-col"),
#             html.Col(className="config-table-col"),
#             html.Col(className="config-table-col"),
#         ]
#         + [HEADER]
#         + [get_row(v) for v in params_subset if not isinstance(v.default, list)],
#     )



import argparse


def get_parser():

    parser = argparse.ArgumentParser(description="Aging of Evolving Genomes in Silico")
    subparsers = parser.add_subparsers(dest="command", help="")

    # subparser_sim
    subparser_sim = subparsers.add_parser("sim", help="run a simulation")
    subparser_sim.add_argument(
        "-c",
        "--config_path",
        type=str,
        help="path to config file",
        default=None,
    )
    subparser_sim.add_argument(
        "-p",
        "--pickle_path",
        type=str,
        help="path to pickle file",
        default=None,
    )
    subparser_sim.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="overwrite old data with new simulation",
        default=None,
    )

    # subparser_gui
    subparser_gui = subparsers.add_parser("gui", help="run GUI")
    subparser_gui.add_argument(
        "--server",
        "-s",
        action="store_true",
        help="run gui – the interactive GUI – in server mode",
        default=False,
    )

    subparser_gui.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="activate the debugger if running locally",
        default=False,
    )

    return parser



import time
import logging

from aegis_sim import variables
from aegis_sim.parameterization import parametermanager


# Decorators
def profile_time(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = (end_time - start_time) * 1000
        logging.info(f"{func.__name__} took {execution_time:.6f} ms to execute")
        return result

    return wrapper


def skip(rate_name) -> bool:
    """Should you skip an action performed at a certain rate"""

    rate = getattr(parametermanager.parameters, rate_name)

    # Skip if rate deactivated
    if rate <= 0:
        return True

    # Do not skip first step
    if variables.steps == 1:
        return False

    # Skip unless step is divisible by rate
    return variables.steps % rate > 0


def steps_to_end() -> int:
    return parametermanager.parameters.STEPS_PER_SIMULATION - variables.steps



import copy
import numpy as np
from aegis_sim.constants import VALID_CAUSES_OF_DEATH
import pathlib
from aegis_sim.utilities.funcs import skip

from aegis_sim.parameterization import parametermanager
from .recorder import Recorder


class FlushRecorder(Recorder):
    """

    Records collections.
    """

    def __init__(self, odir: pathlib.Path):

        self.odir = odir / "gui" / "spectra"
        self.init_odir()

        self.n_ages = parametermanager.parameters.AGE_LIMIT + 1

        self._collection = {
            "age_at_birth": [0] * self.n_ages,
            "additive_age_structure": [0] * self.n_ages,
        }

        self._collection.update({f"age_at_{causeofdeath}": [0] * self.n_ages for causeofdeath in VALID_CAUSES_OF_DEATH})

        self.collection = copy.deepcopy(self._collection)

        self.init_headers()

    def collect(self, key, ages):
        """Add data into memory which will be recorded later."""
        self.collection[key] += np.bincount(ages, minlength=self.n_ages)

    def flush(self):
        """Record data that has been collected over time."""
        # spectra/*.csv | Age distribution of various subpopulations (e.g. population that died of genetic causes)

        if skip("INTERVAL_RATE"):
            return

        for key, val in self.collection.items():
            self.write_age_at(filename=key, collected_values=val)
            # with open(self.odir / f"{key}.csv", "ab") as f:
            #     array = np.array(val)
            #     np.savetxt(f, [array], delimiter=",", fmt="%i")

        # Reinitialize the collection
        self.collection = copy.deepcopy(self._collection)

    def write_age_at(self, filename, collected_values):
        """

        # OUTPUT SPECIFICATION
        path: /gui/spectra/age_at_{cause}.csv
        filetype: csv
        category: demography
        description: Total number of deaths by age and cause of death, within a simulation interval.
        trait granularity: population count
        time granularity: interval
        frequency parameter: INTERVAL_RATE
        structure: An int matrix.
        """
        with open(self.odir / f"{filename}.csv", "ab") as f:
            array = np.array(collected_values)
            np.savetxt(f, [array], delimiter=",", fmt="%i")

    def init_headers(self):
        for key in self._collection.keys():
            with open(self.odir / f"{key}.csv", "ab") as f:
                array = np.arange(self.n_ages)
                np.savetxt(f, [array], delimiter=",", fmt="%i")



import os
import subprocess
import dash
import pathlib
import dash_bootstrap_components as dbc
import platform


def make_button(path):
    return dbc.Button(
        children=[dash.html.I(className="bi bi-folder-symlink-fill"), "Open data location"],
        id={"type": "config-basepath-folder", "index": str(path)},
        value=str(path),
        color="secondary",
        className="me-2",
    )


@dash.callback(
    dash.Output({"type": "config-basepath-folder", "index": dash.MATCH}, "n_clicks"),
    [dash.Input({"type": "config-basepath-folder", "index": dash.MATCH}, "n_clicks")],
    prevent_initial_call=True,
)
def open_file_manager(n_clicks):
    """Open the file manager at the specified path in a platform-independent way."""
    path = pathlib.Path(dash.ctx.triggered_id["index"])

    # TODO does not work on WSL

    if n_clicks > 0:
        system = platform.system().lower()

        if system == "windows":
            os.startfile(path)  # Opens the file manager on Windows
        elif system == "darwin":
            subprocess.run(["open", path])  # Opens the file manager on macOS
        elif system == "linux":
            subprocess.run(["xdg-open", path])  # Opens the file manager on Linux
        elif "microsoft" in os.uname().release.lower():  # For WSL (Windows Subsystem for Linux)
            windows_path = subprocess.check_output(["wslpath", "-w", path]).decode().strip()
            subprocess.run(["explorer.exe", windows_path])  # Opens the Windows Explorer

    return 0  # Reset the click count



from aegis_sim.submodels.reproduction.mutation import Mutator
from aegis_sim.submodels.reproduction.reproduction import Reproducer
from aegis_sim.submodels.abiotic import Abiotic
from aegis_sim.submodels.predation import Predation
from aegis_sim.submodels.resources.starvation import Starvation
from aegis_sim.submodels.infection import Infection
from aegis_sim.submodels.frailty import Frailty
from aegis_sim.submodels.genetics.ploider import Ploider
from aegis_sim.submodels.genetics.architect import Architect
from aegis_sim.utilities.popgenstats import PopgenStats
from aegis_sim.submodels.resources.resources import Resources
from aegis_sim.submodels.reproduction.sexsystem import SexSystem
from aegis_sim.submodels.reproduction.matingmanager import MatingManager


def init(self, parametermanager):

    ##################################
    # INDEPENDENT of other submodels #
    ##################################

    self.abiotic = Abiotic(
        ABIOTIC_HAZARD_SHAPE=parametermanager.parameters.ABIOTIC_HAZARD_SHAPE,
        ABIOTIC_HAZARD_OFFSET=parametermanager.parameters.ABIOTIC_HAZARD_OFFSET,
        ABIOTIC_HAZARD_AMPLITUDE=parametermanager.parameters.ABIOTIC_HAZARD_AMPLITUDE,
        ABIOTIC_HAZARD_PERIOD=parametermanager.parameters.ABIOTIC_HAZARD_PERIOD,
    )
    self.predation = Predation(
        PREDATOR_GROWTH=parametermanager.parameters.PREDATOR_GROWTH,
        PREDATION_RATE=parametermanager.parameters.PREDATION_RATE,
    )
    self.starvation = Starvation(
        STARVATION_RESPONSE=parametermanager.parameters.STARVATION_RESPONSE,
        STARVATION_MAGNITUDE=parametermanager.parameters.STARVATION_MAGNITUDE,
        CLIFF_SURVIVORSHIP=parametermanager.parameters.CLIFF_SURVIVORSHIP,
        CARRYING_CAPACITY=parametermanager.parameters.CARRYING_CAPACITY,
    )
    self.infection = Infection(
        BACKGROUND_INFECTIVITY=parametermanager.parameters.BACKGROUND_INFECTIVITY,
        TRANSMISSIBILITY=parametermanager.parameters.TRANSMISSIBILITY,
        RECOVERY_RATE=parametermanager.parameters.RECOVERY_RATE,
        FATALITY_RATE=parametermanager.parameters.FATALITY_RATE,
    )
    self.frailty = Frailty(
        FRAILTY_MODIFIER=parametermanager.parameters.FRAILTY_MODIFIER,
        AGE_LIMIT=parametermanager.parameters.AGE_LIMIT,
    )

    # Resources
    self.resources = Resources(
        CARRYING_CAPACITY=parametermanager.parameters.CARRYING_CAPACITY,
    )

    # Reproduction
    self.mutator = Mutator(
        MUTATION_RATIO=parametermanager.parameters.MUTATION_RATIO,
        MUTATION_METHOD=parametermanager.parameters.MUTATION_METHOD,
        MUTATION_AGE_MULTIPLIER=parametermanager.parameters.MUTATION_AGE_MULTIPLIER,
    )
    self.sexsystem = SexSystem()
    self.matingmanager = MatingManager()

    # Genetic architecture
    self.ploidy = Ploider(
        REPRODUCTION_MODE=parametermanager.parameters.REPRODUCTION_MODE,
        DOMINANCE_FACTOR=parametermanager.parameters.DOMINANCE_FACTOR,
    )

    # Other
    self.popgenstats = PopgenStats()

    ################################
    # DEPENDENT on other submodels #
    ################################

    # Reproduction
    self.reproduction = Reproducer(
        RECOMBINATION_RATE=parametermanager.parameters.RECOMBINATION_RATE,
        REPRODUCTION_MODE=parametermanager.parameters.REPRODUCTION_MODE,
        mutator=self.mutator,
    )

    # Genetic architecture
    self.architect = Architect(
        ploid=self.ploidy,
        BITS_PER_LOCUS=parametermanager.parameters.BITS_PER_LOCUS,
        PHENOMAP=parametermanager.parameters.PHENOMAP,
        AGE_LIMIT=parametermanager.parameters.AGE_LIMIT,
        THRESHOLD=parametermanager.parameters.THRESHOLD,
        ENVDRIFT_RATE=parametermanager.parameters.ENVDRIFT_RATE,
    )



import numpy as np


class Predation:
    """
    GUI
    Predation is an optional source of mortality, useful for modeling death with prey-predator dynamics.
    [[PREDATION_RATE]] specifies how deadly the predators are; thus if set to 0, no predation deaths will occur.
    Apart from [[PREDATION_RATE]], the probability that an individual actually gets predated depends also on the number of predators; the response curve is logistic.
    The predator population grows according to the logistic Verhulst growth model, whose slope is parameterized by [[PREDATOR_GROWTH]].
    All individuals are equally susceptible to predation; age and genetics have no impact.
    """

    def __init__(self, PREDATOR_GROWTH, PREDATION_RATE):
        self.N = 1
        self.PREDATOR_GROWTH = PREDATOR_GROWTH
        self.PREDATION_RATE = PREDATION_RATE

    def __call__(self, prey_count):

        # TODO Add attrition penalty for when there are no living prey around (but there still are eggs)
        if prey_count == 0:
            return 0

        # Use Verhulst model
        change = self.N * self.PREDATOR_GROWTH * (1 - self.N / prey_count)
        self.N += change

        ratio = np.log(self.N / prey_count)
        # probability to get killed
        # compute logistically
        # multiplication with 2 so that when n_predators == n_prey, the mortality rate is equal to PREDATION_RATE
        fraction_killed = 2 * self.PREDATION_RATE / (1 + np.exp(-ratio))

        return fraction_killed



import yaml
import logging

from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS
from aegis_gui.utilities.utilities import get_config_path, get_container
from aegis_sim.utilities.container import Container
import subprocess

running_processes = {}


def make_config_file(filename, configs):
    configs["PHENOMAP_SPECS"] = []
    configs["NOTES"] = []
    for k, v in configs.items():
        configs[k] = DEFAULT_PARAMETERS[k].convert(v)
    logging.info("Making a config file.")
    config_path = get_config_path(filename)
    with open(config_path, "w") as file_:
        yaml.dump(configs, file_)


def run_simulation(filename, prerun_sim_path):
    global running_processes
    config_path = get_config_path(filename)
    logging.info(f"Running a simulation at path {config_path}.")
    if prerun_sim_path is None:
        pickle_command = []
    else:
        container = Container(prerun_sim_path)
        latest_pickle_path = container.get_path(["pickles"])[-1]
        logging.info(f"Using pickled population from {latest_pickle_path}.")
        pickle_command = ["-p", latest_pickle_path]
    process = subprocess.Popen(
        ["aegis", "sim", "--config_path", config_path] + pickle_command
    )  # used to use sys.executable for cross-platform 'python3' command
    running_processes[filename] = process


def terminate_simulation(simname):
    container = get_container(filename=simname)
    container.terminate()



import dash
import logging
import typing
import dash_bootstrap_components as dbc
from aegis_gui.pages.tab_config import config_input
from aegis_sim.parameterization.parameter import Parameter
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS

from aegis_gui.docs.domains import TEXTS_DOMAIN


def get_pars():
    subsets = {domain: [] for domain in TEXTS_DOMAIN.keys()}
    for param in DEFAULT_PARAMETERS.values():
        if param.domain not in subsets:
            logging.error(f"Parameter domain {param.domain} has no documentation.")
            subsets[param.domain] = []
        subsets[param.domain].append(param)
    return subsets


def get_foldable():

    pars = get_pars()

    elements = [
        dash.html.Div(
            [
                dbc.Button(
                    domain,
                    id={"type": "collapse-button", "index": domain},
                    className="mb-2",
                    color="light",
                    # outline=True,
                    n_clicks=0,
                ),
                dbc.Collapse(
                    dbc.Card(
                        dbc.CardBody(get_line(ps, domain)),
                        # color="light",
                    ),
                    id={"type": "collapse-area", "index": domain},
                    is_open=False,
                ),
            ]
        )
        for domain, ps in pars.items()
    ]

    elements = [
        dash.html.H6("Customize parameters"),
        dbc.Accordion(
            id="config-accordion",
            start_collapsed=True,
            children=[
                dbc.AccordionItem(
                    children=[
                        dbc.Modal(
                            children=[
                                dbc.ModalHeader(dbc.ModalTitle(f"{domain.title()} module")),
                                dbc.ModalBody(TEXTS_DOMAIN.get(domain)),
                                # dbc.ModalFooter(),
                            ],
                            id={"type": "domain-info-modal", "index": domain},
                            className="domain-modal-info",
                        ),
                        dash.html.Div(
                            [
                                "more information at ",
                                dbc.Button(
                                    dash.html.I(className="bi bi-info-square-fill"),
                                    id={"type": "domain-info-modal-trigger", "index": domain},
                                    color="link",
                                    style={"margin": 0, "padding": "0 0 0 0px"},
                                ),
                            ]
                        ),
                        get_line(ps, domain),
                    ],
                    id={"type": "collapse-button", "index": domain},
                    title=dash.html.Div(
                        [
                            domain.title(),
                            dash.html.Span(" "),
                            dbc.Badge(
                                "0 modifications",
                                pill=True,
                                className="ms-1",
                                color="secondary",
                                id={"type": "collapse-badge", "index": domain},
                            ),
                        ],
                        className="position-relative",
                    ),
                )
                for domain, ps in pars.items()
            ],
            # flush=True,
        ),
    ]

    return elements


@dash.callback(
    dash.Output({"type": "domain-info-modal", "index": dash.MATCH}, "is_open"),
    dash.Input({"type": "domain-info-modal-trigger", "index": dash.MATCH}, "n_clicks"),
    prevent_initial_call=True,
)
def toggle_modal(_):
    return True


@dash.callback(
    dash.Output({"type": "collapse-button", "index": dash.ALL}, "active"),
    dash.Output({"type": "collapse-area", "index": dash.ALL}, "is_open"),
    dash.Input({"type": "collapse-button", "index": dash.ALL}, "n_clicks"),
    dash.State({"type": "collapse-area", "index": dash.ALL}, "id"),
    prevent_initial_call=True,
)
def toggle_collapse(n, id_s):
    if n is None:
        return

    triggered_index = dash.ctx.triggered_id["index"]
    is_opens = [id_["index"] == triggered_index for id_ in id_s]

    return is_opens, is_opens
    # if n:
    #     return not is_open
    # return is_open


def get_line(ps: typing.List[Parameter], domain):
    elements = dash.html.Div(
        [
            dash.html.Div(
                [
                    # dbc.Label(p.get_name()),
                    # p.key,
                    # p.dtype.__name__,
                    # (
                    #     dbc.Button(
                    #         dash.html.I(className="bi bi-info-square-fill"),
                    #         id={"type": "info-tooltip", "index": p.key},
                    #         color="link",
                    #     )
                    #     if p.info
                    #     else None
                    # ),
                    # (
                    #     dbc.Tooltip(
                    #         p.info,
                    #         target={"type": "info-tooltip", "index": p.key},
                    #         placement="right",
                    #     )
                    #     if p.info
                    #     else None
                    # ),
                    config_input.get_input_element(param=p),
                ]
            )
            for p in ps
        ],
    )
    return elements



import dash
from dash import html
from aegis_gui.utilities import log

from aegis_gui.pages.tab_simlog import dropdown

# from aegis_gui.pages.tab_simlog.table import generate_initial_simlog


dash.register_page(__name__, path="/simlog", name="AEGIS | Control")


PREFACE = [
    html.Div(
        children=[
            # TODO change text
            html.H1("Simulation log tab"),
            html.P(
                [
                    """Access simulation metadata, download configuration files and simulation data, and terminate or delete simulations."""
                    # """
                    # This is the list tab. Here you can see which simulations you have started and what their status is, including their metadata.
                    # You can also delete simulations and mark which simulations you would like to display in the plot tab.
                    # Note that the 'default' simulation cannot be deleted – it comes with the installation and can be used
                    # as an exploratory example. All parameters for the default simulation have default values.
                    # """,
                ],
                style={"marginBottom": "2rem"},
            ),
        ],
    )
]


def layout(sim=None):
    return html.Div(
        id="simlog-section",
        children=PREFACE + [dropdown.make_select(selected=sim), dropdown.get_info_div()],
        # + [
        #     html.Div(id="simlog-section-table", children=generate_initial_simlog()),
        # table.make_table(),
        # ],
    )



import dash
from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP
from aegis_gui.pages.tab_plot import slider
from aegis_gui.pages.tab_plot import download
from aegis_gui.utilities import log


def get_graph(graph_name):
    info = FIG_SETUP[graph_name]
    graph_div = dash.html.Div(
        children=[
            dash.dcc.Graph(
                # id=figure_id,
                id={"type": "graph-figure", "index": graph_name},
                config={"displayModeBar": False},
                className="figure",
            ),
            dash.html.Div(slider.make_slider(graph_name), style={"width": "100%", "margin": "1rem 0 0 25px"}),
        ],
        id={"type": "graph-div", "index": graph_name},
    )

    return dash.dcc.Loading(
        id={"type": "loading-graph-div", "index": graph_name},
        children=[graph_div],
        overlay_style={"visibility": "visible", "filter": "blur(0px)"},
        parent_style={"display": "none"},
        # delay_show=0,
        # delay_hide=250,
        type="dot",
    )


def get_graph_metadata(graph_name):
    info = FIG_SETUP[graph_name]
    return dash.html.Div(
        [
            dash.html.P(
                children=info["title"],
                className="figure-title",
            ),
            dash.html.Div(
                children=info["description"],
                className="figure-description",
            ),
            download.get_figure_download_button(figure_id=graph_name),
            download.get_figure_download_dcc(figure_id=graph_name),
            download.get_data_download_button(figure_id=graph_name),
            download.get_data_download_dcc(figure_id=graph_name),
        ],
        id={"type": "graph-metadata-div", "index": graph_name},
    )


# Multi Dropdown and Tabs
@dash.callback(
    dash.Output("plot-bottom-right-panel", "children"),
    dash.Input("figure-select", "value"),
)
def update_graph_metadata(figure_selected):
    return get_graph_metadata(figure_selected)


def get_graphs():
    graph_names = [graph_name for graph_name, d in FIG_SETUP.items() if d["supports_multi"] or not d["supports_multi"]]
    graphs = []
    for graph_name in graph_names:
        graph = get_graph(graph_name)
        graphs.append(graph)
    return graphs



"""Overshoot resolver

Decides which individuals to eliminate when there is overcrowding.
"""

import numpy as np


class Starvation:
    """

    GUI
    Starvation is an obligatory source of mortality, useful for modeling death from lack of resources.
    The parameter [[CARRYING_CAPACITY]] specifies the amount of resources.
    Generally, each individual requires one unit of resources; otherwise, they are at risk of starvation.
    When population size exceeds [[CARRYING_CAPACITY]], random individuals will start dying.

    The probability to die is genetics-independent (genetics do not confer protection or susceptibility to starvation).
    However, age can modify the probability to die, depending on the [[STARVATION_RESPONSE]].
    When [[STARVATION_RESPONSE]] is set to treadmill_zoomer, young individuals will start dying first;
    for treadmill_boomer, older individuals die first.

    Under other [[STARVATION_RESPONSE]]s, starvation affects all ages equally, but the dynamics of starvation are different.
    When response is set to gradual, death from starvation is at first low, but increases with each subsequent
    step of insufficient resources (the speed of increase is parameterized by [[STARVATION_MAGNITUDE]]).
    When response is set to treadmill_random, whenever population exceeds the [[CARRYING_CAPACITY]], it is immediately
    and precisely cut down to [[CARRYING_CAPACITY]]. In contrast, when response is set to cliff,
    whenever [[CARRYING_CAPACITY]] is exceeded, the population is cut down to a fraction of the [[CARRYING_CAPACITY]];
    the fraction is specified by the [[CLIFF_SURVIVORSHIP]] parameter.

    Note that if the species is oviparious ([[INCUBATION_PERIOD]]), the produced eggs do not consume resources and are
    immune to starvation mortality (until they hatch).
    """

    def __init__(
        self,
        STARVATION_RESPONSE,
        STARVATION_MAGNITUDE,
        CLIFF_SURVIVORSHIP,
        CARRYING_CAPACITY,
    ):

        self.STARVATION_MAGNITUDE = STARVATION_MAGNITUDE
        self.CLIFF_SURVIVORSHIP = CLIFF_SURVIVORSHIP
        self.CARRYING_CAPACITY = CARRYING_CAPACITY

        self.consecutive_overshoot_n = 0  # For starvation mode

        self.func = {
            "treadmill_random": self._treadmill_random,
            "treadmill_boomer": self._treadmill_boomer,
            "treadmill_zoomer": self._treadmill_zoomer,
            "treadmill_boomer_soft": self._treadmill_boomer_soft,
            "treadmill_zoomer_soft": self._treadmill_zoomer_soft,
            "cliff": self._cliff,
            "gradual": self._gradual,
            "logistic": self._logistic,
        }[STARVATION_RESPONSE]

    def __call__(self, n, resource_availability):
        """Exposed method

        False in `mask` means survival. True in `mask` means death.
        """
        # TODO fix resource availability and carrying capacity logic
        if self.func is not self._gradual:
            resource_availability = self.CARRYING_CAPACITY

        if n <= resource_availability:
            self.consecutive_overshoot_n = 0
            return np.zeros(n, dtype=np.bool_)
        else:
            self.consecutive_overshoot_n += 1
            return self.func(n, resource_availability)

    def _logistic(self, n, resource_availability):
        """Kill random individuals with logistic-like probability."""
        ratio = n / resource_availability

        # when ratio == 1, kill_probability is set to 0
        # when ratio == 2, kill_probability is set to >0
        kill_probability = 2 / (1 + np.exp(-ratio + 1)) - 1

        random_probabilities = np.random.random(n)
        mask = random_probabilities < kill_probability
        return mask

    def _gradual(self, n, resource_availability):
        """Kill random individuals with time-increasing probability.

        The choice of individuals is random.
        The probability of dying increases each consecutive step of overcrowding.
        The probability of dying resets to the base value once the population dips under the maximum allowed size.
        """
        surv_probability = (1 - self.STARVATION_MAGNITUDE) ** self.consecutive_overshoot_n
        random_probabilities = np.random.random(n)
        mask = random_probabilities > surv_probability
        return mask

    def _treadmill_random(self, n, resource_availability):
        """Kill random individuals.

        The population size is brought down to the maximum allowed size in one go.
        """
        indices = np.random.choice(n, n - int(resource_availability), replace=False)
        mask = np.zeros(n, dtype=np.bool_)
        mask[indices] = True
        return mask

    def _cliff(self, n, resource_availability):
        """Kill all individuals except a small random proportion.

        The proportion is defined as the parameter CLIFF_SURVIVORSHIP.
        This function will not necessarily bring the population below the maximum allowed size.
        """
        indices = np.random.choice(
            n,
            int(resource_availability * self.CLIFF_SURVIVORSHIP),
            replace=False,
        )
        mask = np.ones(n, dtype=np.bool_)
        mask[indices] = False
        return mask

    def _treadmill_boomer(self, n, resource_availability):
        """Kill the oldest individuals.

        The population size is brought down to the maximum allowed size in one go.

        NOTE: Why `-resource_availability :`? Because old individuals are at the beginning of the population array.
        """
        mask = np.ones(n, dtype=np.bool_)
        mask[-int(resource_availability) :] = False
        return mask

    def _treadmill_zoomer(self, n, resource_availability):
        """Kill the youngest individuals.

        The population size is brought down to the maximum allowed size in one go.

        NOTE: Why `: resource_availability`? Because young individuals are appended to the end of the population array.
        """
        mask = np.ones(n, dtype=np.bool_)
        mask[: int(resource_availability)] = False
        return mask

    def _treadmill_boomer_soft(self, n, resource_availability):
        """Kill older individuals more.
        Old individuals are positioned more to the front of the array.
        True in the mask means death.
        """
        mask = self._treadmill_soft(1, 0, n, self.CARRYING_CAPACITY)
        return mask

    def _treadmill_zoomer_soft(self, n, resource_availability):
        """Kill younger individuals more.
        Young individuals are positioned later in the array.
        True in the mask means death."""
        mask = self._treadmill_soft(0, 1, n, self.CARRYING_CAPACITY)
        return mask

    @staticmethod
    def _treadmill_soft(linspace_from, linspace_to, n, CARRYING_CAPACITY):
        """
        Young individuals are positioned later in the array; older earlier.
        True in the mask means death.
        """
        p = np.linspace(linspace_from, linspace_to, n) ** 5  # **5 to make it superlinear
        p /= p.sum()  # ensure sum(p) is 1

        mask = np.zeros(n, dtype=np.bool_)

        a = np.arange(n)
        indices_dead = np.random.choice(a, size=n - int(CARRYING_CAPACITY), p=p, replace=False)
        mask[indices_dead] = True
        return mask



"""Modifier of fitness landscape

Modifies topology of fitness landscape over time by changing the interpretation of zeros and ones in genomes.
"""

import numpy as np


class Envdrift:
    """

    GUI
    Environmental drift is deactivated when [[ENVDRIFT_RATE]] is 0.
    Conceptually, environmental drift simulates long-term environmental change such as climate change, resource depletion, pollution, etc.
    The main purpose of environmental drift is to allow the population to keep evolving adaptively.
    When the environment does not change, the fitness landscape is static – initially, the population evolves adaptively as it climbs the fitness landscape but once it approaches the fitness peak,
    natural selection acts mostly to purify new detrimental mutations. When environmental drift is activated, the fitness landscape changes over time – thus, the population keeps evolving adaptively, following the fitness peak.

    """

    def __init__(self, ENVDRIFT_RATE, genome_shape):
        self.ENVDRIFT_RATE = ENVDRIFT_RATE

        if self.ENVDRIFT_RATE == 0:
            self.map = None
        else:
            self.map = np.zeros(genome_shape, dtype=np.bool_)

    def evolve(self, step):
        """Modify the envdrift"""
        if (self.map is None) or (step % self.ENVDRIFT_RATE > 0):
            return

        indices = tuple(np.random.integers(self.map.shape))
        self.map[indices] = ~self.map[indices]

    def call(self, array):
        """Return the genomes reinterpreted"""
        if self.map is not None:
            array = np.logical_xor(self.map, array)
        return array



from aegis_sim import constants


class Trait:
    """Genetic trait

    Contains data on traits encoded in the genome.
    """

    def __init__(self, name, cnf):
        def get(key):
            return getattr(cnf, f"G_{name}_{key}")

        self.name = name

        # Attributes set by the configuration files
        self.evolvable = get("evolvable")
        self.agespecific = get("agespecific")
        self.interpreter = get("interpreter")
        self.lo = get("lo")
        self.hi = get("hi")
        self.initgeno = get("initgeno")
        self.initpheno = get("initpheno")

        # Determine the number of loci encoding the trait
        if self.evolvable:
            if self.agespecific is True:  # one locus per age
                self.length = cnf.AGE_LIMIT
            elif self.agespecific is False:  # one locus for all ages
                self.length = 1
            else:  # custom number of loci
                self.length = self.agespecific
        else:  # no loci for a constant trait
            self.length = 0

        self._validate()

        # Infer positions in the genome
        # self.start = start
        # self.end = self.start + self.length
        # self.slice = slice(self.start, self.end)

        self.start = cnf.AGE_LIMIT * constants.starting_site(self.name)
        self.end = self.start + cnf.AGE_LIMIT
        self.slice = slice(self.start, self.end)

    def _validate(self):
        """Check whether input parameters are legal."""
        if not isinstance(self.evolvable, bool):
            raise TypeError

        if not 0 <= self.initgeno <= 1:
            raise ValueError

        if self.evolvable:
            # if not isinstance(self.agespecific, bool):
            #     raise TypeError

            if self.interpreter not in (
                "uniform",
                "exp",
                "binary",
                "binary_exp",
                "binary_switch",
                "switch",
                "linear",
                "single_bit",
                "const1",
                "threshold",
            ):
                raise ValueError(f"{self.interpreter} is not a valid interpreter type")

            if not 0 <= self.lo <= 1:
                raise ValueError

            if not 0 <= self.hi <= 1:
                raise ValueError

    def __len__(self):
        """Return number of loci used to encode the trait."""
        return self.length

    def __str__(self):
        return self.name



import dash
import dash_bootstrap_components as dbc
from aegis_gui.guisettings.GuiSettings import gui_settings
from aegis_gui.utilities import manipulate, utilities
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS
from .valid_range import is_input_in_valid_range


layout = dash.html.Div(
    [
        dash.html.H6("Run simulation"),
        dash.html.Div(
            id="sim-section-control",
            children=[
                dash.html.Div(
                    children=[
                        dbc.InputGroup(
                            children=[
                                dbc.InputGroupText("Simulation ID"),
                                dbc.Input(
                                    id="config-make-text",
                                    type="text",
                                    placeholder="Enter a unique ID",
                                    # autoComplete="off",
                                    className="me-2",
                                ),
                            ],
                        ),
                        dbc.FormText("", id="config-make-error-text", style={"marginLeft": "8.3rem"}),
                    ],
                    id="config-make-text-group",
                ),
                dbc.Button(
                    [dash.html.I(className="bi bi-rocket-takeoff-fill"), "Launch"],
                    id="simulation-run-button",
                    className="me-1",
                    outline=True,
                    color="success",
                ),
                # dash.html.Button("make config", id="config-make-button"),]
                # dbc.FormFeedback(
                #     "Enter a unique simulation ID",
                #     type="invalid",
                # ),
                # dash.html.P("sssssssssssss", id="simulation-run-text"),
            ],
            style={"display": "flex", "align-items": "flex-start"},
        ),
    ]
)


def decode_config_tab_values(values, ids_):
    for value, id_ in zip(values, ids_):
        param = DEFAULT_PARAMETERS[id_["index"]]
        if param.dtype == bool:
            assert value in {
                "True",
                "False",
                True,
                False,
            }, f"{param.key} value is '{value}' but it should be True or False"
            yield id_["index"], value in {"True", True}
        else:
            yield id_["index"], value


@dash.callback(
    dash.Output({"type": "config-input", "index": dash.ALL}, "value", allow_duplicate=True),
    dash.Input("reset-run-button", "n_clicks"),
    dash.State({"type": "config-input", "index": dash.ALL}, "id"),
    dash.State({"type": "config-input", "index": dash.ALL}, "value"),
    prevent_initial_call=True,
)
def reset_configs(n_clicks, ids, current_values):
    """Why not use DEFAULT_PARAMETERS.get_default_parameters()? To ensure that the order of output values is correct."""

    if n_clicks is None:
        return

    new_values = []

    for id_, current_value in zip(ids, current_values):
        param_name = id_["index"]
        param = DEFAULT_PARAMETERS[param_name]
        new_value = param.default
        if new_value == current_value:
            new_values.append(dash.no_update)
        else:
            new_values.append(new_value)

    return new_values


@dash.callback(
    dash.Output("config-make-text", "value"),
    dash.Input("simulation-run-button", "n_clicks"),
    dash.State("config-make-text", "value"),
    dash.State({"type": "config-input", "index": dash.ALL}, "value"),
    dash.State({"type": "config-input", "index": dash.ALL}, "id"),
    dash.State("prerun-sim-select", "value"),
    prevent_initial_call=True,
)
def click_sim_button(n_clicks, filename, values, ids_, prerun_sim_path):
    """
    Run simulation when sim button clicked (also, implicitly, not disabled).
    """
    if n_clicks is None:
        return

    # make config file
    decoded_pairs = list(decode_config_tab_values(values=values, ids_=ids_))
    input_config = {i: v for i, v in decoded_pairs}
    manipulate.make_config_file(filename, input_config)

    # run simulation
    manipulate.run_simulation(filename, prerun_sim_path=prerun_sim_path)
    return ""


# def is_sim_name_valid(sim_name: str) -> bool:
#     return (
#         (sim_name is not None) and (sim_name != "") and ("." not in sim_name) and (len(sim_name) < MAX_SIM_NAME_LENGTH)
#     )


@dash.callback(
    dash.Output("simulation-run-button", "outline"),
    dash.Output("simulation-run-button", "disabled"),
    dash.Output("config-make-text", "invalid"),
    dash.Output("config-make-text", "valid"),
    dash.Output("config-make-error-text", "children"),
    dash.Input("config-make-text", "value"),
    dash.Input({"type": "config-input", "index": dash.ALL}, "value"),
    dash.State({"type": "config-input", "index": dash.ALL}, "id"),
    dash.State("ticker-store", "data"),
)
def disable_sim_button(filename, values, ids, ticker_store) -> bool:
    """
    Make simulation run button unclickable when any of these is true:
    - input values for parameters are not inside the acceptable server range
    - simulation ID is not valid
    - simulation ID is already used
    """

    block = (True, True, True, False)

    # TODO block here is not correct because the simname is fine
    for value, id_ in zip(values, ids):
        param_name = id_["index"]
        is_value_set = value != "" and value is not None
        if not is_value_set:
            continue

        # check validity of input value
        in_valid_range = is_input_in_valid_range(input_=value, param_name=param_name)
        if not in_valid_range:
            return (
                *block,
                f"Parameter {param_name} is out of valid range",
            )

    # TODO do not use block bc that is not the problem
    currently_running = sum(ticker_store.values())
    if not gui_settings.can_run_more_simulations(currently_running=currently_running):
        return *block, f"Maximum number of simulations running is reached ({currently_running})"

    if filename is None or filename == "":
        return True, True, False, False, ""

    if "." in filename:
        return *block, "ID cannot contain a period (.)"

    if len(filename) > gui_settings.MAX_SIM_NAME_LENGTH:
        return *block, f"ID has more than {gui_settings.MAX_SIM_NAME_LENGTH} characters"

    if utilities.sim_exists(filename):
        return *block, "Entered ID has already been used"

    return False, False, False, True, ""



import dash
from dash import callback, Output, Input, State, MATCH, no_update, dcc, ctx
from aegis_gui.utilities import log
from .valid_range import is_input_in_valid_range
import dash_bootstrap_components as dbc
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS
from aegis_sim.parameterization.parameter import Parameter

# @callback(
#     Output({"type": "config-input", "index": dash.ALL}, "invalid"),
#     Input({"type": "info-badge", "index": dash.ALL}, "children"),
#     Input({"type": "info-badge", "index": dash.ALL}, "children"),
# )
# def handle_config_input(value) -> bool:
#     """
#     Change style of config input so that the user knows that the input value is outside of valid server range.
#     """
#     if ctx.triggered_id is None:
#         return dash.no_update, dash.no_update

#     param_name = ctx.triggered_id["index"]
#     valid = is_input_in_valid_range(input_=value, param_name=param_name)

#     parameter = DEFAULT_PARAMETERS[param_name]
#     value_is_default = value == parameter.default

#     return not valid, "" if value_is_default else "reset"


@callback(
    Output({"type": "collapse-badge", "index": dash.ALL}, "children"),
    Output({"type": "collapse-badge", "index": dash.ALL}, "color"),
    Input({"type": "info-badge", "index": dash.ALL}, "children"),
    State({"type": "info-badge", "index": dash.ALL}, "id"),
    State({"type": "collapse-badge", "index": dash.ALL}, "id"),
    prevent_initial_call=True,
)
def update_badge(badges, id_params, id_domains):
    pnames = [idp["index"] for idp in id_params]
    domains = [idd["index"] for idd in id_domains]
    statuses = {domain: 0 for domain in domains}

    for badge, pname in zip(badges, pnames):
        if badge == "reset":
            domain = DEFAULT_PARAMETERS[pname].domain
            statuses[domain] += 1

    texts = [f"{statuses[domain]} modifications" for domain in domains]
    colors = ["danger" if statuses[domain] > 0 else "secondary" for domain in domains]

    return texts, colors


@callback(
    Output({"type": "config-input", "index": MATCH}, "invalid"),
    Output({"type": "info-badge", "index": MATCH}, "children"),
    Output({"type": "info-badge", "index": MATCH}, "color"),
    Input({"type": "config-input", "index": MATCH}, "value"),
    prevent_initial_call=True,
)
def handle_config_input(value) -> bool:
    """
    Change style of config input so that the user knows that the input value is outside of valid server range.
    """
    if ctx.triggered_id is None:
        return dash.no_update, dash.no_update, dash.no_update

    param_name = ctx.triggered_id["index"]
    valid = is_input_in_valid_range(input_=value, param_name=param_name)

    parameter = DEFAULT_PARAMETERS[param_name]
    value_is_default = value == parameter.default
    new_value = "default" if value_is_default else "reset"
    new_color = "danger" if new_value == "reset" else "secondary"

    return not valid, new_value, new_color


def get_input_element(param: Parameter):
    """
    Generate appropriate Dash input component based on the parameter's data type.
    """

    if not param.show_in_gui:
        return

    common_props = {"id": {"type": "config-input", "index": param.key}, "className": "config-input-class"}

    if param.dtype in [int, float]:
        step = 0 if param.dtype == int else 0.01

        input_element = dbc.Input(
            type="number",
            value=param.default,
            step=step,
            autoComplete="off",
            **common_props,
        )

    elif param.dtype == bool:
        options = ["True", "False"]
        input_element = dbc.Select(
            options=[{"label": k, "value": k} for k in options], value=str(param.default), **common_props
        )

    elif param.dtype == str:
        options = param.drange.strip("{}").split(", ")
        input_element = dbc.Select(
            options=[{"label": k, "value": k} for k in options], value=param.default, **common_props
        )

    else:
        # TODO: resolve dict parameter
        return

    return dbc.InputGroup(
        children=[
            dbc.InputGroupText(
                children=[
                    dash.html.Span(param.get_name(), className="tooltip-underline" if param.info else ""),
                    (
                        dbc.Tooltip(
                            param.info,
                            target={"type": "info-tooltip", "index": param.key},
                            placement="right",
                        )
                        if param.info
                        else None
                    ),
                ],
                id={"type": "info-tooltip", "index": param.key},
            ),
            input_element,
            dbc.InputGroupText(
                [
                    dbc.Badge(
                        "default",
                        pill=True,
                        id={"type": "info-badge", "index": param.key},
                        color="secondary",
                        style={"width": "80px", "cursor": "pointer"},
                    )
                ],
                id={"type": "input-modified-state", "index": param.key},
            ),
        ],
        className="mb-2 mt-2",
    )


@dash.callback(
    Output({"type": "info-badge", "index": dash.MATCH}, "children", allow_duplicate=True),
    Output({"type": "config-input", "index": dash.MATCH}, "value"),
    Input({"type": "info-badge", "index": dash.MATCH}, "n_clicks"),
    prevent_initial_call=True,
)
def reset_config_input(n_clicks):

    if n_clicks is None:
        return dash.no_update, dash.no_update

    param_name = ctx.triggered_id["index"]
    new_value = DEFAULT_PARAMETERS[param_name].default
    return "default", new_value



import dash

from dash import html, dcc
import dash_bootstrap_components as dbc
from aegis_gui.pages.tab_landing import jumbotron, usecases
from aegis_gui.guisettings.GuiSettings import gui_settings

dash.register_page(__name__, name="AEGIS", path="/")


def layout():
    return (
        dbc.Container(
            [
                # html.H1("AEGIS", style={"font-weight": "800", "font-size": "50px"}),
                html.H1("Home tab"),
                html.P(
                    "AEGIS (Aging of Evolving Genomes In Silico) is an advanced software tool that simulates the evolution of life history traits at the genotype-phenotype level. By modeling factors such as resource availability, mortality, mutation rates, and reproductive strategies, AEGIS offers unparalleled insights into the dynamics of life history evolution."
                ),
                jumbotron.jumbotron,
                html.H2("How It Works", className="my-4"),
                html.P(
                    "AEGIS operates by simulating individual organisms within a shared environment. The tool models various life history traits such as survival, reproduction, and mutation rates under different selective pressures. Users can define parameters, run simulations, and analyze outcomes to gain insights into evolutionary processes."
                ),
                html.H2("Use Cases", className="my-4"),
                html.P(
                    "AEGIS is designed to be a tool for theoretical and applied research, as well as a learning aid."
                ),
                usecases.layout,
                html.H2("Concepts modeled", className="my-2"),
                html.Ul(
                    [
                        html.Li("Population dynamics"),
                        html.Li("Starvation"),
                        html.Li("Infection"),
                        html.Li("Predation"),
                        html.Li("Weather mortality"),
                        html.Li("Environmental drift"),
                        html.Li("Pseudogenetic simulation"),
                    ]
                ),
                # html.H2("Contact Us", className="mb-2 mt-4"),
                html.Hr(className="mt-4"),
                html.Div(
                    html.Div(
                        [
                            dbc.Button(
                                [dash.html.I(className="bi bi-envelope-fill"), "Email"],
                                color="secondary",
                                className="me-1",
                                href="mailto:",
                                external_link=True,
                                # outline=True,
                                # style={"backgroundColor": "#1c9be9", "color": "white"},
                            ),
                            dbc.Button(
                                [dash.html.I(className="bi bi-twitter"), "Twitter"],
                                # color="#1c9be9",
                                className="me-1",
                                href="https://x.com/valenzanolab",
                                external_link=True,
                                # outline=True,
                                style={"backgroundColor": "#1c9be9", "color": "white"},
                            ),
                            dbc.Button(
                                [dash.html.I(className="bi bi-github"), "Github"],
                                color="#0e1017",
                                className="me-1",
                                href="https://github.com/valenzano-lab/aegis",
                                external_link=True,
                                # outline=True,
                                style={"backgroundColor": "#0e1017", "color": "white"},
                            ),
                            dbc.Button(
                                [dash.html.I(className="bi bi-globe2"), "Lab website"],
                                color="primary",
                                className="me-1",
                                href="https://valenzano-lab.github.io/labsite/",
                                external_link=True,
                                # outline=True,
                            ),
                        ]
                    ),
                ),
                # html.P(html.Small(f"Version: ...", className="text-secondary")),
            ]
        ),
    )



"""Abstract away genomes."""

import numpy as np


class Genomes:
    def __init__(self, array):
        self.array = array.astype(np.bool_)

    def __len__(self):
        return len(self.array)

    def flatten(self):
        return self.array.reshape(len(self), -1)

    def get(self, individuals):
        return self.array[individuals]

    def add(self, genomes):
        self.array = np.concatenate([self.array, genomes.array])

    def keep(self, individuals):
        self.array = self.array[individuals]

    def get_array(self):
        return self.array.copy()

    def shape(self):
        return self.array.shape

    def __getitem__(self, key):
        new_array = self.array[key]
        return Genomes(new_array)

    # TODO add logicalxor



from aegis_sim.parameterization.parameter import Parameter


def get_default_parameters():
    return {p.key: p.default for p in DEFAULT_PARAMETERS.values()}


def get_species_parameters(SPECIES_PRESET):
    return {p.key: p.presets[SPECIES_PRESET] for p in DEFAULT_PARAMETERS.values() if SPECIES_PRESET in p.presets}


# TODO test these
# TODO value interpolation between ages?
# TODO report invalid values

PRESET_INFO = {
    "human": "One cycle corresponds to 2 years.",
    "mouse": "One cycle corresponds to one month. Source: https://genomics.senescence.info/species/entry.php?species=Mus_musculus",
    "killifish": "One cycle corresponds to one week.",
    "yeast": "",
    "arabidopsis": "",
    "worm": "One cycle corresponds to one day. Up to 300 eggs in optimal conditions.",
    "fruitfly": "One cycle corresponds to one day. Up to 100 eggs per day.",
}

# You need the `key` attribute so you can find the param (in a list, you cannot)
DEFAULT_PARAMETERS = {
    #
    #
    # RECORDING
    "LOGGING_RATE": Parameter(
        key="LOGGING_RATE",
        name="",
        domain="recording",
        default=100,
        info="Frequency of logging (in steps)",
        info_extended="Log files contain information on simulation execution; e.g. errors and speed. 0 for no logging.",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[1, 10000],
    ),
    "TICKER_RATE": Parameter(
        key="TICKER_RATE",
        name="",
        domain="recording",
        default=1,
        info="Frequency of ticking (in seconds)",
        info_extended="Ticker files contain information on simulation status; e.g. running or finished. Once the simulation is finished, it stops updating the ticker file which indicates the time at which the simulation stopped.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[1, 10000],
        show_in_gui=False,
    ),
    "PICKLE_RATE": Parameter(
        key="PICKLE_RATE",
        name="",
        domain="recording",
        default=10000,
        info="Frequency of pickling (in steps)",
        info_extended="0 for no pickles",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        serverrange=lambda x: x >= 1000 or x == 0,
        serverrange_info="0 or [1000, inf)",
        evalrange=[1, 100000],
    ),
    "SNAPSHOT_RATE": Parameter(
        key="SNAPSHOT_RATE",
        name="",
        domain="recording",
        default=10000,
        info="Frequency of recording snapshots (in steps)",
        info_extended="0 for no snapshots",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        serverrange=lambda x: x >= 1000 or x == 0,
        serverrange_info="0 or [1000, inf)",
        evalrange=[1, 10000],
    ),
    "SNAPSHOT_FINAL_COUNT": Parameter(
        key="SNAPSHOT_FINAL_COUNT",
        name="",
        domain="recording",
        default=60,
        info="Number of subsequent snapshots taken at the end of the simulation (in steps)",
        info_extended="0 for no snapshots",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        serverrange=lambda x: 0 <= x <= 60,
        serverrange_info="[0, 60]",
        evalrange=[1, 10],
    ),
    "INTERVAL_RATE": Parameter(
        key="INTERVAL_RATE",
        name="",
        domain="recording",
        default=1000,
        info="Frequency of recording interval data (in steps)",
        info_extended="0 for no gui records",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[1, 10000],
        serverrange=lambda x: x >= 100,
        serverrange_info="[100, inf)",
    ),
    "TE_RATE": Parameter(
        key="TE_RATE",
        name="",
        domain="recording",
        default=10000,
        info="Frequency of starting TE cohorts (in steps)",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[1, 10000],
        serverrange=lambda x: x >= 100,
        serverrange_info="[100, inf)",
    ),
    "TE_DURATION": Parameter(
        key="TE_DURATION",
        name="",
        domain="recording",
        default=500,
        info="Length of tracking TE cohorts (in steps)",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[1, 10000],
    ),
    "POPGENSTATS_RATE": Parameter(
        key="POPGENSTATS_RATE",
        name="",
        domain="recording",
        default=1000,
        info="Frequency of recording population genetic statistics (in steps)",
        info_extended="0 for no recording",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        serverrange=lambda x: x >= 100 or x == 0,
        serverrange_info="0 or [100, inf)",
        evalrange=[1, 10000],
    ),
    "POPGENSTATS_SAMPLE_SIZE": Parameter(
        key="POPGENSTATS_SAMPLE_SIZE",
        name="",
        domain="recording",
        default=100,
        info="Number of individuals to use when calculating population genetic statistics",
        dtype=int,
        drange="{0, [3, inf)}",
        inrange=lambda x: x == 0 or x >= 3,
    ),
    "NOTES": Parameter(
        key="NOTES",
        name="",
        domain="recording",
        default=[],
        info="",
        dtype=list,
        drange="",
    ),
    #
    #
    # STARVATION
    "CARRYING_CAPACITY": Parameter(
        key="CARRYING_CAPACITY",
        name="",
        domain="starvation",
        default=500,
        info="Maximum population size that the environment can sustain",
        info_extended="Starvation mortality is incurred when the population exceeds the carrying capacity.",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        serverrange=lambda x: x <= 5000,
        serverrange_info="[1,5000]",
        evalrange=[1, 1000000],
        previous_keys=["MAX_POPULATION_SIZE"],
    ),
    "CARRYING_CAPACITY_EGGS": Parameter(
        key="CARRYING_CAPACITY_EGGS",
        name="",
        domain="starvation",
        default=500,
        info="Maximum number of eggs that the environment can sustain",
        info_extended="Once the number of eggs exceeds the carrying capacity of eggs, newly laid eggs replace previously laid eggs.",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        serverrange=lambda x: x <= 10000,
        serverrange_info="[1,10000]",
        evalrange=[1, 1000000],
    ),
    "STARVATION_RESPONSE": Parameter(
        key="STARVATION_RESPONSE",
        name="",
        domain="starvation",
        default="gradual",
        info="Mechanism for determining who dies under overcrowding conditions",
        info_extended="The possible modes can differ in the age distribution of mortality and/or the number of individuals removed.",
        dtype=str,
        drange="{gradual, cliff, treadmill_random, treadmill_zoomer, treadmill_boomer, treadmill_boomer_soft, treadmill_zoomer_soft}",
        inrange=lambda x: x
        in (
            "gradual",
            "cliff",
            "treadmill_random",
            "treadmill_zoomer",
            "treadmill_boomer",
            "treadmill_boomer_soft",
            "treadmill_zoomer_soft",
        ),
    ),
    "STARVATION_MAGNITUDE": Parameter(
        key="STARVATION_MAGNITUDE",
        name="",
        domain="starvation",
        default=0.05,
        info="Acceleration of mortality under starvation",
        info_extended="",
        dtype=float,
        drange="(0,1)",
        inrange=lambda x: 0 <= x <= 1,
    ),
    "CLIFF_SURVIVORSHIP": Parameter(
        key="CLIFF_SURVIVORSHIP",
        name="",
        domain="starvation",
        default=None,
        info="Fraction of the population surviving a 'cliff' starvation event",
        info_extended="Modifies the 'cliff' starvation response.",
        dtype=float,
        drange="{None, (0,1)}",
        inrange=lambda x: x is None or (0 < x < 1),
    ),
    #
    #
    # REPRODUCTION
    "INCUBATION_PERIOD": Parameter(
        key="INCUBATION_PERIOD",
        name="",
        domain="reproduction",
        default=0,
        info="Time between fertilization and hatching (in steps)",
        info_extended="0 if egg period is skipped, -1 if hatching occurs only once no living individuals are around.",
        dtype=int,
        drange="[-1, inf)",
        inrange=lambda x: x >= -1,
        presets={},
    ),
    "MATURATION_AGE": Parameter(
        key="MATURATION_AGE",
        name="",
        domain="reproduction",
        default=10,
        info="Minimum age at which an individual can reproduce",
        info_extended="",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        evalrange=[0, 50],
        presets={
            "mouse": 1,  # 1 cycle .. 1 month
        },
    ),
    "REPRODUCTION_ENDPOINT": Parameter(
        key="REPRODUCTION_ENDPOINT",
        name="",
        domain="reproduction",
        default=0,
        info="Minimum age at which an individual can no longer reproduce",
        info_extended="When set to 0, there is no loss of fertility.",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        presets={
            "human": 50,
        },
    ),
    "MAX_OFFSPRING_NUMBER": Parameter(
        key="MAX_OFFSPRING_NUMBER",
        name="",
        domain="reproduction",
        default=1,
        info="Maximum number of offspring that an individual can produce each step.",
        info_extended="Also known as clutch size, litter size or brood size, depending on the species.",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        presets={},
        serverrange=lambda x: x <= 5,
        serverrange_info="[1,5]",
    ),
    # TODO split ploidy from reproduction mode
    "REPRODUCTION_MODE": Parameter(
        key="REPRODUCTION_MODE",
        name="",
        domain="reproduction",
        default="sexual",
        info="Mode of reproduction",
        info_extended="",
        dtype=str,
        drange="{sexual, asexual, asexual_diploid}",
        inrange=lambda x: x in ("sexual", "asexual", "asexual_diploid"),
        presets={
            "yeast": "asexual_diploid",
        },
    ),
    "RECOMBINATION_RATE": Parameter(
        key="RECOMBINATION_RATE",
        name="",
        domain="reproduction",
        default=0.1,
        info="Probability of recombination occuring between two adjacent sites",
        info_extended="If set to 0, there is no recombination.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
        evalrange=[0, 1],
        presets={
            "yeast": 0,
        },
    ),
    "MUTATION_RATIO": Parameter(
        key="MUTATION_RATIO",
        name="",
        domain="reproduction",
        default=0.1,
        info="Ratio of 0->1 mutations to 1->0 mutations",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "MUTATION_AGE_MULTIPLIER": Parameter(
        key="MUTATION_AGE_MULTIPLIER",
        name="",
        domain="reproduction",
        default=0,
        info="Modifier of germline mutation rate",
        info_extended="final germline mutation rate = intrinsic mutation rate + (1 * age * MUTATION_AGE_MULTIPLIER)",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    #
    #
    # GENETICS
    "DOMINANCE_FACTOR": Parameter(
        key="DOMINANCE_FACTOR",
        name="",
        domain="genetics",
        default=1,
        info="Inheritance patterns for non-haploid genomes",
        info_extended="0 for recessive, 0.5 for true additive, 0-1 for partial dominant, 1 for dominant, 1+ for overdominant",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "G_repr_lo": Parameter(
        key="G_repr_lo",
        name="",
        domain="genetics",
        default=0,
        info="Minimum intrinsic fertility",
        dtype=float,
        drange="",
    ),
    "G_repr_hi": Parameter(
        key="G_repr_hi",
        name="",
        domain="genetics",
        default=0.5,
        info="Maximum intrinsic fertility",
        dtype=float,
        drange="",
        evalrange=[0.5, 1],
        # presets={
        #     "mouse": 1,  # 3.5; litter size of 7; 5.4 litters per year; https://genomics.senescence.info/species/entry.php?species=Mus_musculus
        #     "human": 1,  # litter size of 1,
        #     "mouse": 1,  # 5.5; litter size of 5-6
        #     "killifish": 1,  # 50; 1x-1xx eggs, depending on species
        #     "yeast": 1,
        #     "athaliana": 1,  # 1xx seeds per plant
        #     "worm": 1,  # up to 300 eggs in optimal conditions
        #     "fruitfly": 1,  # 100; up to 100 eggs per day in optimal conditions
        # },
    ),
    "G_neut_lo": Parameter(
        key="G_neut_lo",
        name="",
        domain="genetics",
        default=0,
        info="",
        dtype=float,
        drange="",
    ),
    "G_neut_hi": Parameter(
        key="G_neut_hi",
        name="",
        domain="genetics",
        default=1,
        info="",
        dtype=float,
        drange="",
    ),
    "G_muta_lo": Parameter(
        key="G_muta_lo",
        name="",
        domain="genetics",
        default=0,
        info="Minumum intrinsic mutation rate",
        dtype=float,
        drange="",
    ),
    "G_muta_hi": Parameter(
        key="G_muta_hi",
        name="",
        domain="genetics",
        default=1,
        info="Maximum intrinsic mutation rate",
        dtype=float,
        drange="",
    ),
    "G_grow_lo": Parameter(
        key="G_grow_lo",
        name="",
        domain="genetics",
        default=0,
        info="Minimum intrinsic growth rate",
        dtype=float,
        drange="",
    ),
    "G_grow_hi": Parameter(
        key="G_grow_hi",
        name="",
        domain="genetics",
        default=1,
        info="Maximum intrinsic growth rate",
        dtype=float,
        drange="",
    ),
    #
    #
    # ENVIRONMENTAL DRIFT
    "ENVDRIFT_RATE": Parameter(
        key="ENVDRIFT_RATE",
        name="",
        domain="environmental drift",
        default=0,
        info="Frequency of modification to the fitness landscape (in steps)",
        dtype=int,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    #
    #
    # ABIOTIC
    "ABIOTIC_HAZARD_AMPLITUDE": Parameter(
        key="ABIOTIC_HAZARD_AMPLITUDE",
        name="",
        domain="abiotic",
        default=0,
        info="Maximum abiotic hazard",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "ABIOTIC_HAZARD_PERIOD": Parameter(
        key="ABIOTIC_HAZARD_PERIOD",
        name="",
        domain="abiotic",
        default=1,
        info="Period of wave form of abiotic hazard (in steps)",
        dtype=float,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
    ),
    "ABIOTIC_HAZARD_OFFSET": Parameter(
        key="ABIOTIC_HAZARD_OFFSET",
        name="",
        domain="abiotic",
        default=0,
        info="Constant, time-independent abiotic hazard",
        info_extended=r"e.g. 0.01 means that abiotic mortality is increased by 1% each step",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "ABIOTIC_HAZARD_SHAPE": Parameter(
        key="ABIOTIC_HAZARD_SHAPE",
        name="",
        domain="abiotic",
        default="sinusoidal",
        info="Wave form of abiotic hazard",
        dtype=str,
        drange="{sinusoidal, flat, triangle, square, sawtooth, ramp, instant}",
        inrange=lambda x: x in {"sinusoidal", "flat", "triangle", "square", "sawtooth", "ramp", "instant"},
    ),
    #
    #
    # INFECTION
    "BACKGROUND_INFECTIVITY": Parameter(
        key="BACKGROUND_INFECTIVITY",
        name="",
        domain="infection",
        default=0,
        info="Tendency to acquire infection from the environment",
        info_extended="Probability independent of the infection prevalence in the population; thus constant.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "TRANSMISSIBILITY": Parameter(
        key="TRANSMISSIBILITY",
        name="",
        domain="infection",
        default=0,
        info="Tendency to acquire infection from other infected individuals",
        info_extended="Probability dependent on the infection prevalence in the population; thus variable.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "RECOVERY_RATE": Parameter(
        key="RECOVERY_RATE",
        name="",
        domain="infection",
        info="Tendency to transition from infected to healthy status",
        default=0,
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "FATALITY_RATE": Parameter(
        key="FATALITY_RATE",
        name="",
        domain="infection",
        info="Tendency to transition from infected to dead status",
        default=0,
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    #
    #
    # PREDATION
    "PREDATION_RATE": Parameter(
        key="PREDATION_RATE",
        name="",
        domain="predation",
        default=0,
        info="Vulnerability to predators",
        info_extended="Probability to die when number of predators is equal to number of prey. Probability changes logistically with the number of prey.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    "PREDATOR_GROWTH": Parameter(
        key="PREDATOR_GROWTH",
        name="",
        domain="predation",
        default=0,
        info="Intrinsic growth rate of predators",
        info_extended="Growth of the predator population is logistic.",
        dtype=float,
        drange="[0, inf)",
        inrange=lambda x: x >= 0,
    ),
    #
    #
    # GENETIC ARCHITECTURE (COMPOSITE)
    "BITS_PER_LOCUS": Parameter(
        key="BITS_PER_LOCUS",
        name="",
        domain="composite genetic architecture",
        default=8,
        info="Number of bits that each locus has",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        serverrange=lambda x: x <= 10,
        serverrange_info="[1,10]",
        evalrange=[1, 100],
    ),
    "HEADSUP": Parameter(
        key="HEADSUP",
        name="",
        domain="composite genetic architecture",
        default=-1,
        info="-1 if no preevolution, 0 for maturity guarantee, +x for headsup",
        dtype=int,
        drange="{-1, 0, [1, inf)}",
        inrange=lambda x: x in (-1, 0) or x >= 1,
        show_in_gui=False,
    ),
    # "DIFFUSION_FACTOR": Parameter(
    #     key="DIFFUSION_FACTOR",
    #     name="",
    #     domain="composite genetic architecture",
    #     default=1,
    #     info="Window for moving average",
    #     info_extended="When 1, all variants affect one age and trait only. When 1+, they also affect adjacent ages.",
    #     dtype=int,
    #     drange="[1, inf)",
    #     inrange=lambda x: x >= 1,
    #     serverrange=lambda x: x <= 10,
    #     serverrange_info="[1,10]",
    #
    # evalrange=[1, 50],
    # ),
    "G_surv_evolvable": Parameter(
        key="G_surv_evolvable",
        name="",
        domain="composite genetic architecture",
        default=True,
        info="Is survival an evolvable trait?",
        dtype=bool,
        drange="",
        inrange=lambda x: True,
    ),
    "G_surv_agespecific": Parameter(
        key="G_surv_agespecific",
        name="",
        domain="composite genetic architecture",
        default=True,
        info="Is survival age-specific?",
        dtype=bool,
        drange="",
        inrange=lambda x: True,
    ),
    "G_surv_interpreter": Parameter(
        key="G_surv_interpreter",
        name="",
        domain="composite genetic architecture",
        default="binary",
        info="",
        dtype=str,
        drange="",
    ),
    "G_surv_lo": Parameter(
        key="G_surv_lo",
        name="",
        domain="composite genetic architecture",
        default=0,
        info="Minimum survival rate",
        dtype=float,
        drange="",
    ),
    "G_surv_hi": Parameter(
        key="G_surv_hi",
        name="",
        domain="composite genetic architecture",
        default=1,
        info="Maximum survival rate",
        dtype=float,
        drange="",
    ),
    "G_surv_initgeno": Parameter(
        key="G_surv_initgeno",
        name="",
        domain="composite genetic architecture",
        default=1,
        info="Initial survival rate",
        dtype=float,
        drange="",
    ),
    "G_repr_evolvable": Parameter(
        key="G_repr_evolvable",
        name="",
        domain="composite genetic architecture",
        default=True,
        info="Is fertility an evolvable trait?",
        dtype=bool,
        drange="",
    ),
    "G_repr_agespecific": Parameter(
        key="G_repr_agespecific",
        name="",
        domain="composite genetic architecture",
        default=True,
        info="Is fertility age-specific?",
        dtype=bool,
        drange="",
    ),
    "G_repr_interpreter": Parameter(
        key="G_repr_interpreter",
        name="",
        domain="composite genetic architecture",
        default="binary",
        info="",
        dtype=str,
        drange="",
    ),
    "G_repr_initgeno": Parameter(
        key="G_repr_initgeno",
        name="",
        domain="composite genetic architecture",
        default=1,
        info="Initial fertility rate",
        dtype=float,
        drange="",
    ),
    "G_neut_evolvable": Parameter(
        key="G_neut_evolvable",
        name="",
        domain="composite genetic architecture",
        default=False,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_neut_agespecific": Parameter(
        key="G_neut_agespecific",
        name="",
        domain="composite genetic architecture",
        default=False,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_neut_interpreter": Parameter(
        key="G_neut_interpreter",
        name="",
        domain="composite genetic architecture",
        default="binary",
        info="",
        dtype=str,
        drange="",
    ),
    "G_neut_initgeno": Parameter(
        key="G_neut_initgeno",
        name="",
        domain="composite genetic architecture",
        default=1,
        info="",
        dtype=float,
        drange="",
    ),
    "G_muta_evolvable": Parameter(
        key="G_muta_evolvable",
        name="",
        domain="composite genetic architecture",
        default=False,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_muta_agespecific": Parameter(
        key="G_muta_agespecific",
        name="",
        domain="composite genetic architecture",
        default=False,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_muta_interpreter": Parameter(
        key="G_muta_interpreter",
        name="",
        domain="composite genetic architecture",
        default="binary",
        info="",
        dtype=str,
        drange="",
    ),
    "G_muta_initgeno": Parameter(
        key="G_muta_initgeno",
        name="",
        domain="composite genetic architecture",
        default=1,
        info="Initial mutation rate",
        dtype=float,
        drange="",
    ),
    "G_grow_evolvable": Parameter(
        key="G_grow_evolvable",
        name="",
        domain="composite genetic architecture",
        default=True,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_grow_agespecific": Parameter(
        key="G_grow_agespecific",
        name="",
        domain="composite genetic architecture",
        default=False,
        info="",
        dtype=bool,
        drange="",
    ),
    "G_grow_interpreter": Parameter(
        key="G_grow_interpreter",
        name="",
        domain="composite genetic architecture",
        default="binary",
        info="",
        dtype=str,
        drange="",
    ),
    "G_grow_initgeno": Parameter(
        key="G_grow_initgeno",
        name="",
        domain="composite genetic architecture",
        default=0.5,
        info="",
        dtype=float,
        drange="",
    ),
    "THRESHOLD": Parameter(
        key="THRESHOLD",
        name="",
        domain="composite genetic architecture",
        default=None,  # 3
        info="",
        dtype=int,
        drange="",
    ),
    #
    #
    # GENETIC ARCHITECTURE (modifying)
    "PHENOMAP_SPECS": Parameter(
        key="PHENOMAP_SPECS",
        name="",
        domain="modifying genetic architecture",
        default=[],
        info="",
        dtype=list,
        drange="",
        show_in_gui=False,
    ),
    "PHENOMAP": Parameter(
        # TODO is this still needed
        key="PHENOMAP",
        name="",
        domain="modifying genetic architecture",
        default={},
        info="",
        dtype=dict,
        drange="",
        show_in_gui=False,
    ),
    "G_grow_initpheno": Parameter(
        key="G_grow_initpheno",
        name="",
        domain="composite genetic architecture",
        default=0.5,
        info="",
        dtype=float,
        drange="",
    ),
    "G_muta_initpheno": Parameter(
        key="G_muta_initpheno",
        name="",
        domain="modifying genetic architecture",
        default=0.001,
        info="Initial mutation rate",
        dtype=float,
        drange="",
    ),
    "G_surv_initpheno": Parameter(
        key="G_surv_initpheno",
        name="",
        domain="modifying genetic architecture",
        default=1,
        info="Initial survival rate",
        dtype=float,
        drange="",
    ),
    "G_repr_initpheno": Parameter(
        key="G_repr_initpheno",
        name="",
        domain="modifying genetic architecture",
        default=1,
        info="Initial fertility rate",
        dtype=float,
        drange="",
    ),
    "G_neut_initpheno": Parameter(
        key="G_neut_initpheno",
        name="",
        domain="modifying genetic architecture",
        default=1,
        info="",
        dtype=float,
        drange="",
    ),
    #
    #
    # OTHER
    "SPECIES_PRESET": Parameter(
        key="SPECIES_PRESET",
        name="",
        domain="other",
        default=None,
        info="",
        dtype=str,
        drange="None or [" + ",".join(PRESET_INFO.keys()) + "]",
        inrange=lambda x: x in PRESET_INFO.keys() or x is None,
        show_in_docs=False,
        show_in_gui=False,
    ),
    #
    #
    # TIME SCALES
    "STEPS_PER_SIMULATION": Parameter(
        key="STEPS_PER_SIMULATION",
        name="",
        domain="other",
        default=100000,
        info="Number of steps for the simulation to execute",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        serverrange=lambda x: x <= 100000,
        serverrange_info="[1,100000]",
        evalrange=[1, 10000000],
    ),
    "AGE_LIMIT": Parameter(
        key="AGE_LIMIT",
        name="",
        domain="other",
        default=50,
        info="Maximum achievable lifespan (in steps)",
        info_extended="Maximum evolved lifespan is lower than the technically restricted, maximum achievable lifespan.",
        dtype=int,
        drange="[1, inf)",
        inrange=lambda x: x >= 1,
        serverrange=lambda x: x <= 100,
        serverrange_info="[1,100]",
        evalrange=[15, 100],
    ),
    "FRAILTY_MODIFIER": Parameter(
        key="FRAILTY_MODIFIER",
        name="",
        domain="other",
        default=1,
        info="Age-dependent modifier of mortality",
        dtype=float,
        drange="[1, inf)",
        inrange=lambda x: x >= 0,
        serverrange=lambda x: x < 100,
        serverrange_info="[1,100]",
        evalrange=[1, 100],
    ),
    "MORTALITY_ORDER": Parameter(
        key="MORTALITY_ORDER",
        name="",
        domain="other",
        default=["intrinsic", "abiotic", "infection", "predation", "starvation"],
        info="Order in which mortality sources are computed",
        dtype=list,
        drange=None,
        inrange=lambda order: all(
            source in ["intrinsic", "abiotic", "infection", "predation", "starvation"] for source in order
        ),
        show_in_gui=False,
    ),
    #
    #
    # TECHNICAL
    "MUTATION_METHOD": Parameter(
        key="MUTATION_METHOD",
        name="",
        domain="technical",
        default="by_bit",
        info="Vectorized or non-vectorized method of calculating incidence of new mutations",
        info_extended="Mutate by XOR with a randomized bit matrix ('by_bit') or generate random indices to mutate ('by_index')",
        dtype=str,
        drange="{by_bit, by_index}",
        inrange=lambda x: x in ("by_bit", "by_index"),
    ),
    "RANDOM_SEED": Parameter(
        key="RANDOM_SEED",
        name="",
        domain="technical",
        default=None,
        info="Number used as seed for pseudorandom number generator",
        info_extended="If nothing is given, a random integer will be used as the seed; otherwise the given integer will be used as the seed",
        dtype=int,
        drange="{None, (-inf, inf)}",
        inrange=lambda x: True,
    ),
    "PHENOMAP_METHOD": Parameter(
        key="PHENOMAP_METHOD",
        name="",
        domain="technical",
        default="by_loop",
        info="Non-vectorized, vectorized and blank method of calculating phenotypes from genotypes",
        info_extended="Blank method disables pleiotropy.",
        dtype=str,
        drange="{by_loop, by_dot, by_dummy}",
        inrange=lambda x: x in ("by_loop", "by_dot", "by_dummy"),
    ),
}



import dash
from dash import html, dcc
import dash_bootstrap_components as dbc
from aegis_gui.utilities import log, utilities
from aegis_sim.utilities.container import Container
from aegis_gui.pages.tab_plot import select_graph, select_sims, graph, refresh
import aegis_gui.pages.tab_plot.draw  # important callback


dash.register_page(__name__, path="/plot", name="AEGIS | Plot")


def layout(sim=None):  # use function to ensure statelessness

    simnames = [Container(path).basepath.stem for path in utilities.get_sim_paths()]
    dropdown_options = sorted(simnames)

    if not dropdown_options:
        body = [dash.html.P("No simulations to display.")]
    else:
        body = [
            # TODO change text
            html.Div(
                id="plot-top-panel",
                style={"width": "100%"},
                children=[
                    html.Div(
                        [
                            dbc.Row(
                                [
                                    dbc.Col(select_graph.select, width=8),
                                    # dbc.Col(html.Label("(plotting function)"), width=3),
                                ],
                                className="g-0",
                                style={"marginBottom": "0.3rem"},
                            ),
                            dbc.Row(
                                [
                                    dbc.Col(
                                        select_sims.make_multi_dropdown(
                                            dropdown_options=dropdown_options,
                                            dropdown_value=[sim] if sim is not None else [],
                                        ),
                                        width=8,
                                    ),
                                    # dbc.Col(html.Label("(plotted simulations)"), width=3),
                                ],
                                className="g-0",
                            ),
                            dbc.Row(
                                [
                                    dbc.Col(refresh.inputgroup),
                                ]
                            ),
                        ],
                        style={"margin": "2rem 0"},
                    ),
                ],
            ),
            html.Div(
                [
                    html.Div(children=graph.get_graphs(), id="plot-bottom-left-panel", style={"width": "auto"}),
                    html.Div(children=[], id="plot-bottom-right-panel", style={"marginLeft": "2rem"}),
                ],
                style={"width": "100%", "display": "flex"},
            ),
            # reload.make_reload_button(),
            # TODO the order of figures is set by FIG_SETUP; thats how the tabs are generated and how you should return them
            # dropdowns.make_single_dropdown(dropdown_options),
            # tabs.get_tabs_single_layout(),
            # dropdowns.make_multi_dropdown(dropdown_options),
            # tabs.get_tabs_multi_layout(),
        ]

    # TODO plot initially
    children = [html.Div(children=PREFACE + body)]

    return html.Div(id="plot-section", children=children)


PREFACE = [
    html.H1("Plot tab"),
    html.P(
        """
                            Explore simulations and export figures.
                            """
        # This is the plot tab. Here you can explore the simulation visually.
        # You can also download the figures and the data used for plotting.
        # The figures are interactive – if multiple simulations are displayed, you can click on
        # simulation IDs to toggle their visibility; you can also zoom in and out.
        # For figures that show time-specific data, you can use sliders to change the time point plotted.
        # """
    ),
    # dropdowns.make_single_dropdown(dropdown_options),
]



from aegis_sim import constants
from aegis_sim.parameterization.parametermanager import ParameterManager

parametermanager = ParameterManager()

traits = None  # will be redefined below in init_traits


def init_traits(self):
    """
    Here the trait order is hardcoded.
    """
    from aegis_sim.parameterization.trait import Trait

    traits = {}
    for traitname in constants.EVOLVABLE_TRAITS:
        trait = Trait(name=traitname, cnf=parametermanager.parameters)
        traits[traitname] = trait

    self.traits = traits



from aegis_gui import app
from aegis_gui.guisettings.GuiSettings import gui_settings


def run(environment, debug):
    gui_settings.set(environment=environment, debug=debug)
    _app = app.get_app()
    if gui_settings.PORT is None:
        _app.run(debug=gui_settings.DEBUG_MODE)
    else:
        _app.run(debug=gui_settings.DEBUG_MODE, port=gui_settings.PORT)



import logging
import numpy as np
from .recorder import Recorder
from aegis_sim import variables

from aegis_sim.parameterization import parametermanager
# BUG Header duplication; a copy ends up as first row

class TERecorder(Recorder):
    def __init__(self, odir):
        self.odir = odir / "te"
        self.init_odir()
        self.TE_number = 0

    def record(self, T, e):
        """
        Record deaths.
        T .. time / duration (ages)
        E .. event observed (0/alive or 1/dead)

        ###

        To fit this data using lifelines, use this script as inspiration:
            from lifelines import KaplanMeierFitter
            kmf = KaplanMeierFitter()
            te = pd.read_csv("/path/to/te/1.csv")
            kmf.fit(te["T"], te["E"])
            kmf.survival_function_.plot()

        You can compare this to observed survivorship curves:
            analyzer.get_total_survivorship(container).plot()
        """

        assert e in ("alive", "dead")

        step = variables.steps

        # open new file and add header
        if (step % parametermanager.parameters.TE_RATE) == 0 or step == 1:
            data = [["T", "E"]]
            self.write(data, "%s")

        # record deaths
        elif ((step % parametermanager.parameters.TE_RATE) < parametermanager.parameters.TE_DURATION) and e == "dead":
            E = np.repeat(1, len(T))
            data = np.array([T, E]).T
            self.write(data, "%i")

        # flush
        elif (
            ((step % parametermanager.parameters.TE_RATE) == parametermanager.parameters.TE_DURATION)
            or step == parametermanager.parameters.STEPS_PER_SIMULATION
        ) and e == "alive":
            logging.debug(f"Data for survival analysis (T,E) flushed at step {step}.")
            E = np.repeat(0, len(T))
            data = np.array([T, E]).T
            self.write(data, "%i")
            self.TE_number += 1

    def write(self, data, fmt):
        """

        # OUTPUT SPECIFICATION
        path: /te/{te_number}.csv
        filetype: csv
        category: demography
        description: Data for survival analysis; time until event (age at death or current viable age) and the event type (1 if death, 0 if still alive).
        trait granularity: population
        time granularity: interval
        frequency parameter: TE_RATE
        structure: An int matrix.
        """
        path = self.odir / f"{self.TE_number}.csv"
        with open(path, "ab") as file_:
            np.savetxt(file_, data, delimiter=",", fmt=fmt)



from dash import callback, Output, Input, State, ALL, MATCH, ctx, html
from aegis_gui.utilities import log, utilities
import dash_bootstrap_components as dbc
import dash
import os
import shutil


def make(filename):
    return html.Div(
        [
            dbc.Button(
                [html.I(className="bi bi-trash3-fill"), "Delete"],
                id="delete-simulation-button",
                value=filename,
                color="danger",
                className="me-2",
            ),
            dash.dcc.Store("delete-simulation-filename", data=filename),
        ]
    )


# @callback(
#     Output({"type": "delete-simulation-button", "index": MATCH}, "disabled"),
#     Input({"type": "delete-simulation-button", "index": MATCH}, "n_clicks"),
# )
#
# def disable_delete_simulation_button(n_clicks):
#     if n_clicks is None:
#         return False
#     return True


@callback(
    Output("simlog-section-table", "children"),
    Input({"type": "delete-simulation-button", "index": ALL}, "n_clicks"),
    State("simlog-section-table", "children"),
    prevent_initial_call=True,
)
def change_simlog(n_clicks, current):

    # If delete button triggered the action, delete the simulation
    if isinstance(ctx.triggered_id, dict) and ctx.triggered_id.get("type") == "delete-simulation-button":
        filename = ctx.triggered_id["index"]
        config_path = utilities.get_config_path(filename)
        sim_path = config_path.parent / filename
        remove_simulation_data(sim_path=sim_path, config_path=config_path)
        return current
    else:
        raise Exception


def delete_simulation(filename):
    config_path = utilities.get_config_path(filename)
    sim_path = config_path.parent / filename
    remove_simulation_data(sim_path=sim_path, config_path=config_path)


def remove_simulation_data(sim_path, config_path):
    """Removes the simulation directory and configuration file."""
    # Remove the directory and all its contents
    if os.path.isdir(sim_path):
        shutil.rmtree(sim_path)  # This removes the directory and all its contents

    # Remove the configuration file
    if os.path.isfile(config_path):
        os.remove(config_path)  # This removes the file


@callback(
    Output("modal-centered", "is_open"),
    dash.Output("sim-select", "options"),
    dash.Output("sim-select", "value"),
    Input("delete-simulation-button", "n_clicks"),
    Input("close-centered", "n_clicks"),
    Input("permanently-delete", "n_clicks"),
    State("delete-simulation-filename", "data"),
    dash.State("sim-select", "options"),
    dash.State("one-click-deletion", "value"),
    prevent_initial_call=True,
)
def toggle_modal(n1, n2, n3, filename, current_options, one_click_deletion):

    if ctx.triggered_id == "delete-simulation-button" and n1 is not None:
        if one_click_deletion:
            # repeating myself below
            delete_simulation(filename=filename)
            new_options = [d for d in current_options if d["label"] != filename]
            return False, new_options, new_options[0]["value"] if new_options else ""
        return True, dash.no_update, dash.no_update

    if ctx.triggered_id == "close-centered" and n2 is not None:
        return False, dash.no_update, dash.no_update

    if ctx.triggered_id == "permanently-delete" and n3 is not None:
        delete_simulation(filename=filename)
        new_options = [d for d in current_options if d["label"] != filename]
        return False, new_options, new_options[0]["value"] if new_options else ""

    # TODO issue when no new options

    return dash.no_update, dash.no_update, dash.no_update


def make_modal():
    return dbc.Modal(
        [
            dbc.ModalHeader(dbc.ModalTitle("Delete simulation"), close_button=True),
            dbc.ModalBody("Are you sure you want to permanently delete the simulation data?"),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        "Yes, delete",
                        id="permanently-delete",
                        n_clicks=0,
                        color="danger",
                        outline=True,
                    ),
                    dbc.Button(
                        "Close",
                        id="close-centered",
                        n_clicks=0,
                    ),
                ]
            ),
        ],
        id="modal-centered",
        centered=True,
        is_open=False,
    )



"""Interpreter of genomic loci

Transforms bool array into an array of numbers.
These numbers can be loosely understood as gene activity.
"""

import numpy as np
from aegis_sim import parameterization


class Interpreter:
    def __init__(self, BITS_PER_LOCUS, THRESHOLD):

        self.BITS_PER_LOCUS = BITS_PER_LOCUS
        self.THRESHOLD = THRESHOLD

        self.exp_base = 0.5  # Important for _exp
        self.binary_exp_base = 0.98  # Important for _binary_exp

        # Parameters for the binary interpreter
        binary_weights = 2 ** np.arange(self.BITS_PER_LOCUS)[::-1]
        self.binary_weights = binary_weights / binary_weights.sum()

        # Parameters for the binary switch interpreter
        binary_switch_weights = 2 ** np.arange(self.BITS_PER_LOCUS)[::-1]
        binary_switch_weights[-1] = 0  # Switch bit does not add to locus value
        self.binary_switch_weights = binary_switch_weights / binary_switch_weights.sum()
        # e.g. when BITS_PER_LOCUS is 4, binary_switch_weights are [4/7, 2/7, 1/7, 0]

        # Parameters for the linear interpreter
        linear_weights = np.arange(self.BITS_PER_LOCUS)[::-1] + 1
        self.linear_weights = linear_weights / linear_weights.sum()

    def call(self, loci, interpreter_kind):
        """Exposed method"""

        # shape is (n_individuals, age_limit, bits_per_locus)
        assert loci.shape[0] > 0
        assert loci.shape[1] == parameterization.parametermanager.parameters.AGE_LIMIT
        assert loci.shape[2] == self.BITS_PER_LOCUS

        method = {
            "const1": self._const1,
            "single_bit": self._single_bit,
            "threshold": self._threshold,
            "linear": self._linear,
            "binary": self._binary,
            "switch": self._switch,
            "binary_switch": self._binary_switch,
            "uniform": self._uniform,
            "exp": self._exp,
            "binary_exp": self._binary_exp,
        }[interpreter_kind]
        interpretome = method(loci)
        return interpretome

    @staticmethod
    def _const1(loci):
        return np.ones((len(loci), 1))

    @staticmethod
    def _single_bit(loci):
        return loci[:, :, 0]

    def _threshold(self, loci):
        """Penna interpreter
        Cares only about the first bit of the locus.
        """
        return (~loci[:, :, 0]).cumsum(-1) < self.THRESHOLD

    def _linear(self, loci):
        return np.matmul(loci, self.linear_weights)

    def _binary(self, loci):
        """Interpret locus as a binary number and normalize.

        High resolution (can produce 2^bits_per_locus different numbers).
        Position-dependent.
        """

        return np.matmul(loci, self.binary_weights)

    @staticmethod
    def _switch(loci):
        """Return 0 if all bits are 0; 1 if all bits are 1; 0 or 1 randomly otherwise.

        Low resolution (can produce 2 different numbers).
        Position-independent.
        """
        sums = loci.mean(2)
        rand_values = np.random.random(loci.shape[:-1]) < 0.5
        return np.select([sums == 0, (sums > 0) & (sums < 1), sums == 1], [0, rand_values, 1])

    def _binary_switch(self, loci):
        """Interpret first n-1 bits as a binary number if the last bit is 1.

        High resolution (can produce 2^(bits_per_locus-1) different numbers).
        Position-dependent.
        """
        where_on = loci[:, :, -1] == 1  # Loci which are turned on
        values = np.zeros(loci.shape[:-1], dtype=np.float32)  # Initialize output array with zeros
        values[where_on] = loci[where_on].dot(
            self.binary_switch_weights
        )  # If the locus is turned on, make the value in the output array be the binary value
        return values

    def _uniform(loci):
        """Return normalized sum of bits.

        Medium resolution (can produce bits_per_locus+1 different numbers).
        Position-independent.
        """
        return loci.sum(-1) / loci.shape[-1]

    def _exp(self, loci):
        """Return base^total_number_of_zeros.

        Medium resolution (can produce bits_per_locus+1 different numbers).
        Suitable for generating very small numbers.
        Position-independent.
        """
        return self.exp_base ** np.sum(1 - loci, axis=2)

    def _binary_exp(self, loci):
        """Return base^binary_value_of_locus

        High resolution (can produce 2^bits_per_locus different numbers).
        Suitable for generating very small numbers.
        Position-dependent.
        """
        binary = self._binary(loci)
        return self.binary_exp_base**binary



import logging
import pathlib

from .recorder import Recorder
from aegis_sim import variables

from aegis_sim.parameterization import parametermanager
from aegis_sim.utilities.funcs import skip

class PickleRecorder(Recorder):
    def __init__(self, odir: pathlib.Path):
        self.odir = odir / "pickles"
        self.init_odir()

    def write(self, population):
        """

        # OUTPUT SPECIFICATION
        path: /pickles/{step}
        filetype: pickle
        category: log
        description: A file that records the Population class instance which can be used to seed a future simulation.
        trait granularity: individual
        time granularity: snapshot
        frequency parameter: PICKLE_RATE
        structure: Binary python file.
        """

        step = variables.steps
        should_skip = skip("PICKLE_RATE")
        is_first_step = step == 1
        is_last_step = step == parametermanager.parameters.STEPS_PER_SIMULATION

        if is_first_step or not should_skip or is_last_step:
            logging.debug(f"pickle recorded at step {step}.")
            path = self.odir / str(step)
            population.save_pickle_to(path)



class Ploider:
    """ """

    def __init__(self, REPRODUCTION_MODE, DOMINANCE_FACTOR):
        self.REPRODUCTION_MODE = REPRODUCTION_MODE
        self.DOMINANCE_FACTOR = DOMINANCE_FACTOR
        self.y = self.get_ploidy(REPRODUCTION_MODE)

    @staticmethod
    def get_ploidy(REPRODUCTION_MODE):
        return {
            "sexual": 2,
            "asexual": 1,
            "asexual_diploid": 2,
        }[REPRODUCTION_MODE]

    def diploid_to_haploid(self, loci):
        """Merge two arrays encoding two chromatids into one array.

        Arguments:
            loci: A bool numpy array with shape (population size, ploidy, genome length, BITS_PER_LOCUS)

        Returns:
            A bool numpy array with shape (population size, genome length, BITS_PER_LOCUS)
        """

        assert len(loci.shape) == 4, len(loci.shape)  # e.g. (45, 2, 250, 8)
        assert loci.shape[1] in (1, 2), loci.shape[1]  # ploidy

        # TODO handle polyploidy too
        # compute homozygous (0, 1) or heterozygous (0.5)
        zygosity = loci.mean(1)
        mask = zygosity == 0.5

        # correct heterozygous with DOMINANCE_FACTOR
        if isinstance(self.DOMINANCE_FACTOR, list):
            zygosity[mask] = self.DOMINANCE_FACTOR[mask]
        else:
            zygosity[mask] = self.DOMINANCE_FACTOR

        assert len(zygosity.shape) == 3, len(zygosity.shape)

        return zygosity



import dash
import re
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS

from aegis_sim.recording.recordingmanager import RecordingManager
from aegis_sim.submodels.reproduction.reproduction import Reproducer
from aegis_sim.submodels import predation, infection, abiotic
from aegis_sim.submodels.resources import starvation
from aegis_sim.submodels.genetics.composite.architecture import CompositeArchitecture
from aegis_sim.submodels.genetics.modifying.architecture import ModifyingArchitecture
from aegis_sim.submodels.genetics.envdrift import Envdrift


def extract_gui_from_docstring(class_):
    docstring = class_.__doc__
    gui_section = docstring.split("GUI")[1]
    # text = replace_params_in_brackets_with_span(gui_section)
    parsed = parse_gui_docstrings(gui_section)
    return list(parsed)


def parse_gui_docstrings(text):
    pattern = r"(\[\[|\]\])"
    parts = re.split(pattern, text)
    is_parameter = False
    for part in parts:
        if part == "[[":
            is_parameter = True
        elif part == "]]":
            is_parameter = False
        else:
            if is_parameter:
                yield get_parameter_span(part)
            else:
                yield dash.html.Span(part)


def highlight_terms_in_markdown(class_):
    docstring = class_.__doc__
    gui_section = docstring.split("GUI")[1]
    text = gui_section
    if not isinstance(text, str):
        text = str(text)

    def replace_highlight(match):
        return f"**{match.group(1)}**"

    pattern = r"\[\[(.*?)\]\]"
    highlighted_text = re.sub(pattern, replace_highlight, text)
    highlighted_text = highlighted_text.replace("\n\n    ", "\n    ")
    highlighted_text = highlighted_text.replace("\n    ", "\n")
    return highlighted_text


def get_parameter_span(name):
    reformatted_name = name.replace("_", " ").lower()
    param = DEFAULT_PARAMETERS[name]
    info = param.info
    return dash.html.Span(
        reformatted_name,
        title=info,
        # style={"font-style": "italic"},
    )


genetics_text = """
GUI
Every individual carries their own genome. In AEGIS, those are bit strings (arrays of 0's and 1's), passed on from parent to offspring, and mutated in the process.
The submodel genetics transforms genomes into phenotypes; more specifically – into intrinsic phenotypes – biological potentials to exhibit a certain trait (e.g. probability to reproduce).
These potentials are either realized or not realized, depending on the environment (e.g. availability of resources), interaction with other individuals (e.g. availability of mates) and interaction with other traits (e.g. survival).

In AEGIS, genetics is simplified in comparison to the biological reality – it references no real genes and it simulates no molecular interactions; thus, it cannot be used to answer questions about specific genes, metabolic pathways or molecular mechanisms.
However, in AEGIS, in comparison to empirical datasets, genes are fully functionally characterized (in terms of their impact on the phenotype), and are to be studied as functional, heritable genetic elements – in particular, their evolutionary dynamics.

The configuration of genetics – the genetic architecture – is highly flexible. This includes specifying which traits are evolvable number of genetic elements (i.e. size of genome)...
AEGIS offers two genetic architectures – composite and modifying. They are mutually exclusive and are described in detail below...
"""


def get_texts(func):
    return {
        "starvation": func(starvation.Starvation),
        "predation": func(predation.Predation),
        "infection": func(infection.Infection),
        "abiotic": func(abiotic.Abiotic),
        "reproduction": func(Reproducer),
        "recording": func(RecordingManager),
        "genetics": genetics_text,
        "composite genetic architecture": func(CompositeArchitecture),
        "modifying genetic architecture": func(ModifyingArchitecture),
        "environmental drift": func(Envdrift),
        "technical": "",  # TODO add
        "other": "",  # TODO add
    }


TEXTS_DOMAIN = get_texts(extract_gui_from_docstring)
TEXTS_DOMAIN_MD = get_texts(highlight_terms_in_markdown)



"""
This script is executed when AEGIS is imported (`import aegis`). Execute functions by running `aegis.run_from_{}`.
AEGIS can be started in multiple ways; each of these functions starts AEGIS from a different context. 
"""

import logging
import pathlib

import aegis_gui
import aegis_sim
from aegis.log import set_logging
from aegis.parse import get_parser


def start_from_terminal():
    parser = get_parser()
    args = parser.parse_args()
    set_logging(level=logging.DEBUG)

    if args.command == "sim":
        config_path = pathlib.Path(args.config_path).absolute() if args.config_path else None
        pickle_path = pathlib.Path(args.pickle_path).absolute() if args.pickle_path else None
        aegis_sim.run(
            custom_config_path=config_path,
            pickle_path=pickle_path,
            overwrite=args.overwrite,
            custom_input_params={},
        )
    elif args.command == "gui":
        if args.server:
            aegis_gui.run(environment="server", debug=False)
        else:
            aegis_gui.run(environment="local", debug=args.debug)
    else:
        parser.print_help()



import re
import psutil
import logging
from typing import Set


def run_ps_af() -> Set[str]:
    """Returns names of simulations whose processes are currently running based on the OS."""
    config_files = set()
    try:
        # Define the pattern for matching process command lines
        pattern = re.compile(r"python3 -m aegis --config_path .*/(.*)\.yml")

        # Iterate through all running processes
        for proc in psutil.process_iter(["cmdline"]):
            try:
                # Get the command line arguments of the process
                cmdline = " ".join(proc.info["cmdline"])

                # Match the command line with the pattern
                match = pattern.search(cmdline)
                if match:
                    config_file = match.group(1)
                    config_files.add(config_file)
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                # Process has been terminated or we don't have permission to access it
                continue

    except Exception as e:
        logging.error(f"An error occurred: {e}")

    return config_files



import dash_bootstrap_components as dbc
from dash import html
from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP

buttons = html.Div(
    [
        dbc.Button("Regular", color="primary", className="me-1"),
        dbc.Button("Active", color="primary", active=True, className="me-1"),
        dbc.Button("Disabled", color="primary", disabled=True),
    ]
)

gns = [graph_name for graph_name, d in FIG_SETUP.items() if d["supports_multi"] or not d["supports_multi"]]

initial_graph = "death table"
assert initial_graph in gns

select = html.Div(
    [
        dbc.InputGroup(
            [
                dbc.InputGroupText("Plotting function"),
                dbc.Select(
                    id="figure-select",
                    options=[{"label": gn, "value": gn} for gn in gns],
                    value=initial_graph,
                    className="plot-dropdown",
                    persistence=True,
                ),
            ]
        ),
    ]
)



import numpy as np
import logging

from aegis_sim import variables
from aegis_sim import submodels
from aegis_sim.constants import VALID_CAUSES_OF_DEATH
from aegis_sim.dataclasses.population import Population
from aegis_sim.recording import recordingmanager
from aegis_sim.parameterization import parametermanager


class Bioreactor:
    def __init__(self, population: Population):
        self.eggs: Population = None
        self.population: Population = population

    ##############
    # MAIN LOGIC #
    ##############

    def run_step(self):
        """Perform one step of simulation."""

        # If extinct (no living individuals nor eggs left), do nothing
        if len(self) == 0:
            logging.debug("Population went extinct.")
            recordingmanager.summaryrecorder.extinct = True
        # Mortality sources
        self.mortalities()

        recordingmanager.popsizerecorder.write_before_reproduction(self.population)
        self.growth()  # size increase
        self.reproduction()  # reproduction
        self.age()  # age increment and potentially death
        self.hatch()
        submodels.architect.envdrift.evolve(step=variables.steps)
        submodels.resources.replenish()

        # Record data
        recordingmanager.popsizerecorder.write_after_reproduction(self.population)
        recordingmanager.flushrecorder.collect("additive_age_structure", self.population.ages)  # population census
        recordingmanager.picklerecorder.write(self.population)
        recordingmanager.featherrecorder.write(self.population)
        recordingmanager.guirecorder.record(self.population)
        recordingmanager.flushrecorder.flush()
        recordingmanager.popgenstatsrecorder.write(
            self.population.genomes, self.population.phenotypes.extract(ages=self.population.ages, trait_name="muta")
        )  # TODO defers calculation of mutation rates; hacky
        recordingmanager.summaryrecorder.record_memuse()
        recordingmanager.terecorder.record(self.population.ages, "alive")

    ###############
    # STEP LOGIC #
    ###############

    def mortalities(self):
        for source in parametermanager.parameters.MORTALITY_ORDER:
            if source == "intrinsic":
                self.mortality_intrinsic()
            elif source == "abiotic":
                self.mortality_abiotic()
            elif source == "infection":
                self.mortality_infection()
            elif source == "predation":
                self.mortality_predation()
            elif source == "starvation":
                self.mortality_starvation()
            else:
                raise ValueError(f"Invalid source of mortality '{source}'")

    def mortality_intrinsic(self):
        probs_surv = self.population.phenotypes.extract(ages=self.population.ages, trait_name="surv")
        age_hazard = submodels.frailty.modify(hazard=1 - probs_surv, ages=self.population.ages)
        mask_kill = np.random.random(len(probs_surv)) < age_hazard
        self._kill(mask_kill=mask_kill, causeofdeath="intrinsic")

    def mortality_abiotic(self):
        hazard = submodels.abiotic(variables.steps)
        age_hazard = submodels.frailty.modify(hazard=hazard, ages=self.population.ages)
        mask_kill = np.random.random(len(self.population)) < age_hazard
        self._kill(mask_kill=mask_kill, causeofdeath="abiotic")

    def mortality_infection(self):
        submodels.infection(self.population)
        # TODO add age hazard
        mask_kill = self.population.infection == -1
        self._kill(mask_kill=mask_kill, causeofdeath="infection")

    def mortality_predation(self):
        probs_kill = submodels.predation(len(self))
        # TODO add age hazard
        mask_kill = np.random.random(len(self)) < probs_kill
        self._kill(mask_kill=mask_kill, causeofdeath="predation")

    def mortality_starvation(self):
        resource_availability = submodels.resources.scavenge(np.ones(len(self.population)))
        # TODO add age hazard
        mask_kill = submodels.starvation(
            n=len(self.population),
            resource_availability=resource_availability.sum(),
        )
        self._kill(mask_kill=mask_kill, causeofdeath="starvation")

    def reproduction(self):
        """Generate offspring of reproducing individuals.
        Initial is set to 0.
        """

        # Check if fertile
        mask_fertile = (
            self.population.ages >= parametermanager.parameters.MATURATION_AGE
        )  # Check if mature; mature if survived MATURATION_AGE full cycles
        if parametermanager.parameters.REPRODUCTION_ENDPOINT > 0:
            mask_menopausal = (
                self.population.ages >= parametermanager.parameters.REPRODUCTION_ENDPOINT
            )  # Check if menopausal; menopausal when lived through REPRODUCTION_ENDPOINT full cycles
            mask_fertile = (mask_fertile) & (~mask_menopausal)

        if not any(mask_fertile):
            return

        # Check if reproducing
        probs_repr = self.population.phenotypes.extract(ages=self.population.ages, trait_name="repr", part=mask_fertile)

        # Binomial calculation
        n = parametermanager.parameters.MAX_OFFSPRING_NUMBER
        p = probs_repr
        num_repr = np.random.binomial(n=n, p=p)
        mask_repr = num_repr > 0

        if sum(num_repr) == 0:
            return

        # Indices of reproducing individuals
        who = np.repeat(np.arange(len(self.population)), num_repr)

        # Count ages at reproduction
        ages_repr = self.population.ages[who]
        recordingmanager.flushrecorder.collect("age_at_birth", ages_repr)

        # Increase births statistics
        self.population.births += num_repr

        # Generate offspring genomes
        parental_genomes = self.population.genomes.get(individuals=who)
        parental_sexes = self.population.sexes[who]

        muta_prob = self.population.phenotypes.extract(ages=self.population.ages, trait_name="muta", part=mask_repr)[
            mask_repr
        ]
        muta_prob = np.repeat(muta_prob, num_repr[mask_repr])

        offspring_genomes = submodels.reproduction.generate_offspring_genomes(
            genomes=parental_genomes,
            muta_prob=muta_prob,
            ages=ages_repr,
            parental_sexes=parental_sexes,
        )
        offspring_sexes = submodels.sexsystem.get_sex(len(offspring_genomes))

        # Randomize order of newly laid egg attributes ..
        # .. because the order will affect their probability to be removed because of limited carrying capacity
        order = np.arange(len(offspring_sexes))
        np.random.shuffle(order)
        offspring_genomes = offspring_genomes[order]
        offspring_sexes = offspring_sexes[order]

        # Make eggs
        eggs = Population.make_eggs(
            offspring_genomes=offspring_genomes,
            step=variables.steps,
            offspring_sexes=offspring_sexes,
            parental_generations=np.zeros(len(offspring_sexes)),  # TODO replace with working calculation
        )
        if self.eggs is None:
            self.eggs = eggs
        else:
            self.eggs += eggs
        if len(self.eggs) > parametermanager.parameters.CARRYING_CAPACITY_EGGS:
            indices = np.arange(len(self.eggs))[-parametermanager.parameters.CARRYING_CAPACITY_EGGS :]
            # TODO biased
            self.eggs *= indices

    def growth(self):
        max_growth_potential = self.population.phenotypes.extract(ages=self.population.ages, trait_name="grow")
        gathered_resources = submodels.resources.scavenge(max_growth_potential)
        self.population.sizes += gathered_resources

    def age(self):
        """Increase age of all by one and kill those that surpass age limit.
        Age denotes the number of full cycles that an individual survived and reproduced.
        AGE_LIMIT is the maximum number of full cycles an individual can go through.
        """
        self.population.ages += 1
        mask_kill = self.population.ages >= parametermanager.parameters.AGE_LIMIT
        self._kill(mask_kill=mask_kill, causeofdeath="age_limit")

    def hatch(self):
        """Turn eggs into living individuals"""

        # If nothing to hatch
        if self.eggs is None:
            return

        # If something to hatch
        if (
            (
                parametermanager.parameters.INCUBATION_PERIOD == -1 and len(self.population) == 0
            )  # hatch when everyone dead
            or (parametermanager.parameters.INCUBATION_PERIOD == 0)  # hatch immediately
            or (
                parametermanager.parameters.INCUBATION_PERIOD > 0
                and variables.steps % parametermanager.parameters.INCUBATION_PERIOD == 0
            )  # hatch with delay
        ):
            self.eggs.phenotypes = submodels.architect.__call__(self.eggs.genomes)
            self.population += self.eggs
            self.eggs = None

    ################
    # HELPER FUNCS #
    ################

    def _kill(self, mask_kill, causeofdeath):
        """Kill individuals and record their data."""

        assert causeofdeath in VALID_CAUSES_OF_DEATH

        # Skip if no one to kill
        if not any(mask_kill):
            return

        # Count ages at death
        # if causeofdeath != "age_limit":
        ages_death = self.population.ages[mask_kill]
        recordingmanager.flushrecorder.collect(f"age_at_{causeofdeath}", ages_death)
        recordingmanager.terecorder.record(ages_death, "dead")

        # Retain survivors
        self.population *= ~mask_kill

    def __len__(self):
        """Return the number of living individuals and saved eggs."""
        return len(self.population) + len(self.eggs) if self.eggs is not None else len(self.population)



import dash
from dash import dcc, Input, callback, Output, MATCH, State
from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP
from aegis_gui.pages.tab_plot.plot import prep_x, gen_fig
from aegis_gui.utilities import log


def make_slider(figure_id):

    needs_slider = lambda figure_id: FIG_SETUP[figure_id]["prep_x"] == prep_x.get_ages

    return dcc.Slider(
        id={"type": "graph-slider", "index": figure_id},
        className="" if needs_slider(figure_id) else "invisible",
        min=0,
        max=1,
        step=1,
        value=0,
        marks=None,
        tooltip={},
        updatemode="drag",
    )


# def needs_slider(fig_name):
#     return FIG_SETUP[fig_name]["prep_x"] == prep_x.get_ages


# Single Dropdown and Tabs
@callback(
    Output({"type": "graph-figure", "index": MATCH}, "figure", allow_duplicate=True),
    Output({"type": "graph-slider", "index": MATCH}, "max", allow_duplicate=True),  # updates slider max as you use it
    Input({"type": "graph-slider", "index": MATCH}, "value"),
    State({"type": "graph-slider", "index": MATCH}, "id"),
    State("dropdown-multi", "value"),
    dash.Input("color-mode-switch", "value"),
    # State("dropdown-single", "value"),
    prevent_initial_call=True,
)
@log.log_debug
def update(slider_value, slider_id, dropdown_multi_value, dark_mode):
    fig_name = slider_id["index"]
    # supports_multi = FIG_SETUP[fig_name]["supports_multi"]
    # dropdown_values = dropdown_multi_value if supports_multi else [dropdown_single_value]
    dropdown_values = dropdown_multi_value
    fig, max_iloc = gen_fig.generate_figure(fig_name, dropdown_values, iloc=slider_value, dark_mode=dark_mode)
    return fig, max_iloc



def get_sorted_allele_frequencies(interval_genotypes):
    total_frequency = interval_genotypes.sum(0)
    return interval_genotypes.T.assign(total=total_frequency).sort_values(by="total", ascending=False).T.iloc[:-1]


def get_derived_allele_freq(interval_genotypes):
    reference = interval_genotypes.round()
    derived_allele_freq = (
        interval_genotypes.iloc[1:].reset_index(drop=True) - reference.iloc[:-1].reset_index(drop=True)
    ).abs()
    return derived_allele_freq


def get_mean_allele_freq(interval_genotypes):
    mean_allele_freq = interval_genotypes.mean(0)
    return mean_allele_freq


def get_quantile_allele_freq(interval_genotypes, quantile):
    quantile_allele_freq = interval_genotypes.quantile(quantile)
    return quantile_allele_freq



import dash
import dash_bootstrap_components as dbc
from aegis_gui.guisettings.GuiSettings import gui_settings

left_jumbotron = dbc.Col(
    dash.html.Div(
        [
            dash.html.H2("Simulate", className="display-3"),
            dash.html.Hr(className="my-2"),
            dash.html.P("Jump right in and run evolutionary simulations of life history."),
            dash.dcc.Link(dbc.Button("Start", color="light", outline=True), href=gui_settings.wrap_href("config")),
        ],
        className="h-100 p-5 text-white bg-primary rounded-3",
    ),
    md=6,
)

right_jumbotron = dbc.Col(
    dash.html.Div(
        [
            dash.html.H2("Analyze", className="display-3"),
            dash.html.Hr(className="my-2"),
            dash.html.P("Or, inspect results of pre-run simulations."),
            dash.dcc.Link(
                dbc.Button(
                    "Show me plots",
                    color="light",
                    outline=True,
                    id="read-jumbotron-button",
                ),
                href=gui_settings.wrap_href("plot"),
            ),
        ],
        className="h-100 p-5 bg-secondary text-white border rounded-3",
    ),
    md=6,
)

jumbotron = dbc.Row(
    [left_jumbotron, right_jumbotron],
    className="align-items-md-stretch",
)



import numpy as np


def init(self, custom_config_path, pickle_path, RANDOM_SEED):
    self.steps = 1
    self.custom_config_path = custom_config_path
    self.pickle_path = pickle_path
    self.random_seed = np.random.randint(1, 10**6) if RANDOM_SEED is None else RANDOM_SEED
    np.random.seed(self.random_seed)



EVOLVABLE_TRAITS = ("surv", "repr", "muta", "neut", "grow")

TRAIT_N = len(EVOLVABLE_TRAITS)


def starting_site(trait_name):
    return EVOLVABLE_TRAITS.index(trait_name)


VALID_CAUSES_OF_DEATH = (
    "intrinsic",
    "abiotic",
    "infection",
    "predation",
    "starvation",
    "age_limit",
)



import pandas as pd
import logging
import pathlib

from .recorder import Recorder
from aegis_sim import submodels

class PhenomapRecorder(Recorder):
    """

    Records once.
    """

    def __init__(self, odir: pathlib.Path):
        self.odir = odir
        self.init_odir()

    def write(self):
        """

        # OUTPUT SPECIFICATION
        path: /phenomap.csv
        filetype: csv
        category: genotype
        description: A static list of phenotypic effects of each genomic site. 
        trait granularity: N/A
        time granularity: N/A
        frequency parameter: once
        structure: A table with four columns: effector site index, affected trait, affected age, effect magnitude. Each row represents an effect of a single site on a specific trait expressed at a specific age.
        """
        architecture = submodels.architect.architecture

        if hasattr(architecture, "phenomap"):
            phenolist = architecture.phenomap.phenolist
            pd.DataFrame(phenolist).to_csv(
                self.odir / "phenomap.csv", index=None, header=["bit", "trait", "age", "weight"]
            )
        else:
            logging.info("Phenomap is empty.")
